# -*- coding: utf-8 -*-

"""
.. module:: quality_cim_quality.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.quality.cim_quality document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_display_name,
        _set_description,
        _set_summary_fields
        )


def _set_display_name(ctx):
    """Sets document display name.

    """
    name = ""
    if ctx.meta.project:
        name += ctx.meta.project.upper()
        name += " "
    name += "Dataset Quality Control Report"

    ctx.ext.display_name = ctx.ext.full_display_name = name


def _set_description(ctx):
    """Sets document description.

    """
    if len(ctx.doc.reports) and ctx.doc.reports[0].measure is not None:
        ctx.ext.description = ctx.doc.reports[0].measure.name


def _set_summary_fields(ctx):
    """Sets document summary fields.

    """
    if len(ctx.meta.external_ids):
        ctx.ext.summary_fields += (
            ctx.meta.external_ids[0].value,
        )





# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.codecs.dict.encoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encodes a term to a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.cv.model import Term
from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Collection
from pyesdoc.cv.model import Scope



def encode(instance):
    """Encodes an instance of a domain model class as a dictionary.

    :param pyesdoc.cv.Entity instance: A domain model class instance to be encoded as a dictionary.

    :returns: Instance encoded as a simple dictionary.
    :rtype: dict

    :raises TypeError: If instance is a domain model class instance.

    """
    try:
        _ENCODERS[type(instance)]
    except KeyError:
        raise TypeError("Type encoding unsupported: {}".format(type(instance)))

    obj = _ENCODERS[type(instance)](instance)
    obj['_type'] = unicode(instance.__module__)

    return obj


def _encode_authority(instance):
    """Encodes a term authority as a dictionary.

    """
    obj = dict()
    obj['description'] = instance.description
    obj['label'] = instance.label
    obj['name'] = instance.name
    obj['scopes'] = [_encode_scope(i) for i in instance.scopes]
    obj['url'] = instance.url

    return obj


def _encode_scope(instance):
    """Encodes a term scope as a dictionary.

    """
    obj = dict()
    obj['collections'] = [_encode_collection(i) for i in instance.collections]
    obj['description'] = instance.description
    obj['idx'] = instance.idx
    obj['label'] = instance.label
    obj['name'] = instance.name
    obj['uid'] = instance.uid
    obj['url'] = instance.url

    return obj


def _encode_collection(instance):
    """Encodes a term collection as a dictionary.

    """
    obj = dict()
    obj['create_date'] = instance.create_date
    obj['description'] = instance.description
    obj['idx'] = instance.idx
    obj['label'] = instance.label
    obj['name'] = instance.name
    obj['terms'] = ["{}:{}".format(i.name, i.uid) for i in instance.terms]
    obj['uid'] = instance.uid
    obj['url'] = instance.url

    return obj


def _encode_term(instance):
    """Encodes a term as a dictionary.

    """
    obj = instance.__dict__.copy()
    del obj['collection']
    del obj['io_path']
    obj['parent'] = None if instance.parent is None else instance.parent.uid
    obj['associations'] = [i.uid for i in instance.associations]

    return obj


# Map of supported types to encoding functions.
_ENCODERS = {
    Authority: _encode_authority,
    Collection: _encode_collection,
    Scope: _encode_scope,
    Term: _encode_term,
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.exceptions
   :platform: Unix, Windows
   :synopsis: Library exceptions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

# Unsupported schema error.
class UnsupportedOntologySchema(ValueError):
    pass


# -*- coding: utf-8 -*-

"""
.. module:: test_types.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Set of document types to be tested.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import test_type_cim_v1_activity_ensemble_1
import test_type_cim_v1_activity_ensemble_2
import test_type_cim_v1_activity_numerical_experiment
import test_type_cim_v1_activity_simulation_run
import test_type_cim_v1_data_data_object_1
import test_type_cim_v1_data_data_object_2
import test_type_cim_v1_grids_gridspec
import test_type_cim_v1_misc_document_set
import test_type_cim_v1_quality_cim_quality
import test_type_cim_v1_shared_platform
import test_type_cim_v1_software_model_component_1
import test_type_cim_v1_software_model_component_2
import test_type_cim_v1_software_statistical_model_component


# Set of test modules.
MODULES = (
    test_type_cim_v1_activity_ensemble_1,
    test_type_cim_v1_activity_ensemble_2,
    test_type_cim_v1_activity_numerical_experiment,
    test_type_cim_v1_activity_simulation_run,
    test_type_cim_v1_data_data_object_1,
    test_type_cim_v1_data_data_object_2,
    test_type_cim_v1_grids_gridspec,
    test_type_cim_v1_misc_document_set,
    test_type_cim_v1_quality_cim_quality,
    test_type_cim_v1_shared_platform,
    test_type_cim_v1_software_model_component_1,
    test_type_cim_v1_software_model_component_2,
    # test_type_cim_v1_software_statistical_model_component,
)


# Set of stateful test module fields.
STATE_FIELDS = (
    "DOC_TYPE",
    "DOC_TYPE_KEY",
    "DOC_FILE",
    "DOC_UID",
    "DOC_VERSION",
    "DOC_DATE",
    "DOC_PROJECT",
    "DOC_INSTITUTE",
    "DOC_AUTHOR"
    )


# Initial state of test modules.
INITIAL_STATE = {mod:{} for mod in MODULES}
for mod in MODULES:
    for field in STATE_FIELDS:
        INITIAL_STATE[mod][field] = getattr(mod, field)


def reset_all():
    """Resets all module state prior to usage.

    """
    for mod in MODULES:
        reset(mod)


def reset(mod):
    """Resets module state prior to usage.

    :param module mod: Test module.

    """
    assert mod in MODULES
    assert mod in INITIAL_STATE
    for field in STATE_FIELDS:
        setattr(mod, field, INITIAL_STATE[mod][field])



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean upper / lower boundaries'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel:free_surface'] = {
    'description': 'Properties of free surface in ocean',
    'properties': [
        ('scheme', 'ENUM:free_surface_types', '1.1',
            'Free surface scheme in ocean'),
        ('embeded_seaice', 'bool', '1.1',
            'Is the sea-ice embeded in the ocean model (instead of levitating) ?'),
        ]
    }

DETAILS['toplevel:bottom_boundary_layer'] = {
    'description': 'Properties of bottom boundary layer in ocean',
    'properties': [
        ('type_of_bbl', 'ENUM:bottom_bl_types', '1.1',
            'Type of bottom boundary layer in ocean'),
        ('lateral_mixing_coef', 'int', '0.1',
            'If bottom BL is diffusive, specify value of lateral mixing coefficient (in m2/s)'),
        ('sill_overflow', 'str', '1.1',
            'Describe any specific treatment of sill overflows')
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['free_surface_types'] = {
    'description': 'Type of free surface in ocean',
    'is_open': True,
    'members': [
        ('Linear implicit', None),
        ('Linear filtered', None),
        ('Linear semi-explicit', None),
        ('Non-linear implicit', None),
        ('Non-linear filtered', None),
        ('Non-linear semi-explicit', None)
        ]
    }

ENUMERATIONS['bottom_bl_types'] = {
    'description': 'Type of bottom boundary layer in ocean',
    'is_open': True,
    'members': [
        ('Diffusive', None),
        ('Acvective', None)
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.io.read.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Reads an authority's vocabularies from file sytem.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import glob
import os

from pyesdoc.cv.codecs import from_json


# Manifest file name.
_MANIFEST = "MANIFEST"


def read_authority(dpath):
    """Reads authority CV data from file system.

    :param str dpath: Path to directory to which a CV hierarchy has been written.

    :returns: Authority CV data.
    :rtype: pyesdoc.cv.Authority

    """
    if not os.path.isdir(dpath):
        raise OSError("Invalid directory.")
    if not os.path.isfile(os.path.join(dpath, _MANIFEST)):
        raise OSError("Invalid MANIFEST.")

    # Read authority from manifest.
    fpath = os.path.join(dpath, _MANIFEST)
    with open(fpath, "r") as fstream:
        authority = from_json(fstream.read())
        authority.io_path = fpath

    # Read terms.
    for scope in authority:
        for collection in scope:
            collection.terms = []
            collection.terms += _read_terms(dpath, scope, collection)

    # Set inter-concept hierachy.
    terms = {}
    for scope in authority:
        scope.authority = authority
        for collection in scope:
            collection.scope = scope
            for term in collection:
                term.collection = collection
                terms[term.uid] = term

    # Set inter-term hierarchies.
    for term in terms.values():
        if term.parent in terms:
            term.parent = terms[term.parent]

    # Set intra-term hierarchies.
    for term in [i for i in terms.values() if i.associations]:
        term.associations = [terms[i] if i in terms else i for i in term.associations]

    return authority


def _read_terms(dpath, scope, collection):
    """Reads terms from file system.

    """
    dpath = os.path.join(dpath, scope.name)
    dpath = os.path.join(dpath, collection.name)
    dpath = os.path.join(dpath, "*")

    return [_read_term(i) for i in glob.iglob(dpath)]


def _read_term(fpath):
    """Reads terms from file system.

    """
    # Decode term from JSON file.
    with open(fpath, "r") as fstream:
        term = from_json(fstream.read())
    term.io_path = fpath

    return term



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.class_validator
   :platform: Unix, Windows
   :synopsis: Validates ontology class definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import re

from pyesdoc.mp.ontologies.schemas import utils



# Regular expressions.
_RE_CLASS_NAME = '^[a-z_0-9]+$'
_RE_CLASS_PROPERTY_NAME = '^[a-z_]+$'
_RE_CLASS_PROPERTY_TYPE = '^[a-z_]+\.?[a-z_]+$'
_RE_CLASS_REFERENCE = '^[a-z_]+\.?[a-z_]+$'

# Whitelist of valid class property cardinality.
_CLASS_PROPERTY_CARDINALITIES = {'0.1', '0.N', '1.1', '1.N'}

# Whitelist of valid class property simple types.
_SIMPLE_CLASS_PROPERTY_TYPES = {
    'bool',
    'date',
    'datetime',
    'float',
    'int',
    'str',
    'unicode',
    'uri',
    'uuid'
}


def _validate_class_property_name(ctx, module, factory, name):
    """Validates a class property name.

    """
    if not re.match(_RE_CLASS_PROPERTY_NAME, name):
        err = 'Invalid property: {0}.[{1}] --> name format must be lower_case_underscore'
        err = err.format(ctx.get_name(factory, module), name)
        ctx.set_error(err)


def _validate_class_property_type(ctx, module, factory, name, typeof):
    """Validates a class property type.

    """
    typeof, qualifier = utils.parse_type(typeof)

    if typeof in _SIMPLE_CLASS_PROPERTY_TYPES:
        return

    if not re.match(_RE_CLASS_PROPERTY_TYPE, typeof):
        err = 'Invalid property: {0}.[{1}] --> '
        err += 'type format must be lower_case_underscore '
        err += '(for class references a "." is expected)'
        err = err.format(ctx.get_name(factory, module), name)
        ctx.set_error(err)

    if qualifier and not re.match(_RE_CLASS_PROPERTY_TYPE, qualifier):
        err = 'Invalid property qualifier: {0}.[{1}] --> '
        err += 'type format must be lower_case_underscore '
        err += '(for class references a "." is expected)'
        err = err.format(ctx.get_name(factory, module), name)
        ctx.set_error(err)

    if len(typeof.split('.')) == 1:
        err = 'Invalid property: {0}.[{1}] --> simple type must be in {2}'
        err = err.format(ctx.get_name(factory, module), name, _SIMPLE_CLASS_PROPERTY_TYPES)
        ctx.set_error(err)


def _validate_class_property_cardinality(ctx, module, factory, name, cardinality):
    """Validates a class property cardinality.

    """
    if cardinality not in _CLASS_PROPERTY_CARDINALITIES:
        err = 'Invalid property: {0}.[{1}] --> cardinality must be in {2}'
        err = err.format(ctx.get_name(factory, module), name, _CLASS_PROPERTY_CARDINALITIES)
        ctx.set_error(err)


def _validate_class_property(ctx, module, factory, name, typeof, cardinality):
    """Validates a class property.

    """
    _validate_class_property_name(ctx, module, factory, name)
    _validate_class_property_type(ctx, module, factory, name, typeof)
    _validate_class_property_cardinality(ctx, module, factory, name, cardinality)


def validate(ctx, module, factory, cls):
    """Validates a class definition.

    """
    if 'base' not in cls:
        err = 'Invalid class: {} --> required attribute "base" is missing'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if 'is_abstract' not in cls:
        err = 'Invalid class: {} --> required attribute "is_abstract" is missing'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if not re.match(_RE_CLASS_NAME, factory.__name__):
        err = 'Invalid class: {} --> name format must be lower_case_underscore'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if cls['base'] is not None and not re.match(_RE_CLASS_REFERENCE, cls['base']):
        err = 'Invalid class: {0} --> base class reference [{1}] format must be lower_case_underscore'
        err = err.format(ctx.get_name(factory, module), cls['base'])
        ctx.set_error(err)

    if cls['base'] is not None and not len(cls['base'].split('.')) == 2:
        err = 'Invalid class: {} --> base class reference must contain a "." spliting the package and type references, e.g. activity.numerical_activity'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if not isinstance(cls['is_abstract'], bool):
        err = 'Invalid class: {} --> "is_abstract" attribute must be a boolean'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if 'doc_strings' in cls and not isinstance(cls['doc_strings'], dict):
        err = 'Invalid class: {} --> "doc_strings" attribute must be a dict'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)

    if 'properties' in cls and not isinstance(cls['properties'], list):
        err = 'Invalid class: {} --> "properties" attribute must be a list'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if [p for p in cls.get('properties', []) if not isinstance(p, tuple) or not len(p) in {3, 4}]:
        err = 'Invalid class: {} --> all properties must be tuples (name, type, cardinality, doc_string)'
        err = err.format(ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    for p in cls.get('properties', []):
        _validate_class_property(ctx, module, factory, p[0], p[1], p[2])



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.ontologies.type_info.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates type information related to registered ontologies.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from collections import defaultdict



# Set of registered ontologies.
ONTOLOGIES = ()

# Set of registered packages.
PACKAGES = ()

# Set of registered classes.
CLASSES = ()

# Class properties.
CLASS_PROPERTIES = {}

# Class own properties.
CLASS_OWN_PROPERTIES = {}

# Base classes.
BASE_CLASSES = defaultdict(tuple)

# Classes with base classes.
BASE_CLASSED = ()

# Sub classes.
SUB_CLASSES = defaultdict(tuple)

# Classes that have been sub classed.
SUB_CLASSED = ()

# Set of registered enums.
ENUMS = ()

# Set of registered types.
TYPES = ()

# Help text.
DOC_STRINGS = {}

# Map of types to keys.
KEYS = {}

# Map of types to constraints.
CONSTRAINTS = {}

# Set of registered document types.
DOCUMENT_TYPES = ()


def register(o):
    """Registers an ontology for use.

    :param o: Ontology being registered.
    :type o: module

    """
    global BASE_CLASSED
    global SUB_CLASSED
    global ONTOLOGIES
    global PACKAGES
    global TYPES
    global DOCUMENT_TYPES
    global CLASSES
    global ENUMS

    if o in ONTOLOGIES:
        return

    ONTOLOGIES += (o,)
    for v in [v.type_info for v in o.VERSIONS]:
        PACKAGES += v.PACKAGES
        TYPES += v.TYPES
        DOCUMENT_TYPES += v.DOCUMENT_TYPES
        CLASSES += v.CLASSES
        CLASS_PROPERTIES.update(v.CLASS_PROPERTIES)
        CLASS_OWN_PROPERTIES.update(v.CLASS_OWN_PROPERTIES)
        BASE_CLASSES.update(v.BASE_CLASSES)
        BASE_CLASSED += v.BASE_CLASSED
        SUB_CLASSES.update(v.SUB_CLASSES)
        SUB_CLASSED += v.SUB_CLASSED
        ENUMS += v.ENUMS
        KEYS.update(v.KEYS)
        CONSTRAINTS.update(v.CONSTRAINTS)
        DOC_STRINGS.update(v.DOC_STRINGS)


def _get_type(target):
    """Returns type from target.

    """
    if target in TYPES:
        return target
    return type(target)


def get_own_properties(kls):
    """Returns set of class own properties.

    :param kls: A class or class instance within one of the registered ontologies.

    :returns: Set of a class's own properties.
    :rtype: set

    """
    kls = _get_type(kls)

    return set(CLASS_OWN_PROPERTIES.get(kls, []))


def get_properties(kls):
    """Returns set of class properties.

    :param kls: A class or class instance within one of the registered ontologies.

    :returns: Set of class properties.
    :rtype: set

    """
    kls = _get_type(kls)

    return set(CLASS_PROPERTIES.get(kls, []))


def get_is_document_type(kls):
    """Returns flag indicating whether a class is a document type.

    :param kls: A class or class instance within one of the registered ontologies.

    :returns: True if class is considered to be a "document".
    :rtype: bool

    """
    kls = _get_type(kls)

    return kls in DOCUMENT_TYPES


def get_sub_classes(kls):
    """Returns set of sub-classes.

    """
    kls = _get_type(kls)

    return [KEYS[i] for i in SUB_CLASSES[kls]]


def get_is_sub_classed(kls):
    """Returns flag indicating whether a class is sub-classed or not.

    :param kls: A class or class instance within one of the registered ontologies.

    :returns: True if class is sub-classed.
    :rtype: bool

    """
    kls = _get_type(kls)

    return kls in SUB_CLASSED


def get_bases(kls):
    """Returns set of base classes.

    """
    kls = _get_type(kls)

    return [KEYS[i] for i in BASE_CLASSES[kls]]


def get_has_base_class(kls):
    """Returns flag indicating whether a class has a base class or not.

    :param kls: A class or class instance within one of the registered ontologies.

    :returns: True if class has a base class.
    :rtype: bool

    """
    kls = _get_type(kls)

    return kls in BASE_CLASSED


def get_doc_string(typeof, attribute=None):
    """Returns a documentation string.

    :param typeof: A type within one of the registered ontologies.
    :param attribute: Name of either a class property or an enum member.

    :returns: A documentation string.
    :rtype: str

    """
    typeof = _get_type(typeof)

    if attribute:
        return DOC_STRINGS.get((typeof, attribute), "").strip()
    return DOC_STRINGS.get(typeof, "").strip()


def get_key(typeof, attribute=None):
    """Returns a type's key.

    :param typeof: A type within one of the registered ontologies.
    :param attribute: Name of either a class property or an enum member.

    :returns: A key.
    :rtype: str

    """
    typeof = _get_type(typeof)

    if attribute:
        return KEYS.get((typeof, attribute), "").strip()
    return KEYS.get(typeof, "").strip()


def get_constraints(typeof, attribute=None):
    """Returns a type's constraints.

    :param typeof: A type within one of the registered ontologies.
    :param attribute: Name of either a class property or an enum member.

    :returns: A key.
    :rtype: str

    """
    typeof = _get_type(typeof)

    if attribute:
        return set(CONSTRAINTS.get((typeof, attribute), ""))

    return {p: get_constraints(typeof, p) for p in get_properties(typeof)}


# Auto register cim.
from pyesdoc.ontologies import cim
register(cim)


# Remove import dross.
del defaultdict
del cim



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.mp.ontologies.generators.qconfig.qconfig_generator.py
   :platform: Unix, Windows
   :synopsis: Generates qconfig file for an ES-DOC Questionnaire Ontology

.. moduleauthor:: Allyn Treshansky <allyn.treshansky@colorado.edu>

"""


from pyesdoc.mp.ontologies.generators import generator_utils as gu
from pyesdoc.mp.ontologies.generators.generator import Generator
from pyesdoc.mp.ontologies.generators.qconfig import utils as qgu

from collections import OrderedDict
from errno import EEXIST as PATH_EXISTS
import datetime
import json
import os


class QConfigGenerator(Generator):

    def on_start(self, ctx):
        """Event handler for the start parse event.

        :param GeneratorContext ctx: Generation context information.

        """

        qconfig = OrderedDict()
        ctx.qconfig = qconfig

    def on_end(self, ctx):
        """Event handler for the end parse event.

        :param GeneratorContext ctx: Generation context information.

        """

        qconfig = ctx.qconfig
        output_filepath = qgu.get_ontology_path(ctx)
        if not os.path.exists(os.path.dirname(output_filepath)):
            try:
                os.makedirs(os.path.dirname(output_filepath))
            except OSError as error:  # handle race condition, as per http://stackoverflow.com/a/12517490
                if error.errno != PATH_EXISTS:
                    raise error
        with open(output_filepath, 'w') as f:
            json.dump(qconfig, f)
        f.closed

    def on_ontology_parse(self, ctx):
        """Event handler for the ontology parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ontology = ctx.ontology
        qconfig = ctx.qconfig

        qconfig["name"] = ontology.name
        qconfig["version"] = ontology.version
        qconfig["documentation"] = ontology.doc_string
        qconfig["ontology_type"] = "SCHEMA"
        # qconfig["ontology_base"] = no base for schemas
        qconfig["date"] = str(datetime.datetime.now())
        qconfig["classes"] = {
            "inherited": [],
            "excluded": [],
            "defined": [],
        }

    def on_class_parse(self, ctx):
        """Event handler for the class parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        cls = ctx.cls
        qconfig = ctx.qconfig
        qconfig_classes = qconfig["classes"]["defined"]

        if not cls.is_abstract:

            new_qconfig_class = OrderedDict()
            new_qconfig_class["name"] = cls.name
            new_qconfig_class["package"] = cls.package.name
            new_qconfig_class["documentation"] = cls.doc_string
            new_qconfig_class["is_document"] = cls.is_entity
            new_qconfig_class["is_meta"] = cls.is_document_meta

            class_id = getattr(cls, "id", None)
            if class_id:
                new_qconfig_class["id"] = class_id

            label = cls.pstr
            if label:
                new_qconfig_class["label"] = OrderedDict()
                new_qconfig_class["label"]["text"] = label.text
                new_qconfig_class["label"]["fields"] = []
                for field in label.fields:
                    new_qconfig_class["label"]["fields"].append(field)

            new_qconfig_class["categories"] = {
                "inherited": [],
                "excluded": [],
                "defined": [],
            }
            new_qconfig_class["properties"] = {
                "inherited": [],
                "excluded": [],
                "defined": [],
            }
            for prop in cls.all_properties:

                new_qconfig_property = OrderedDict()
                new_qconfig_property["name"] = prop.name
                # new_qconfig_property["package"] = prop.package.name
                new_qconfig_property["is_meta"] = prop.is_linked_document
                new_qconfig_property["documentation"] = prop.doc_string
                new_qconfig_property["cardinality"] = prop.cardinality
                new_qconfig_property["is_nillable"] = not prop.is_required

                property_id = getattr(prop, "id", None)
                if property_id:
                    new_qconfig_property["id"] = property_id

                property_type = qgu.get_property_type(prop)
                new_qconfig_property["property_type"] = property_type
                if property_type == qgu.QCONFIG_ATOMIC_PROPERTY_TYPE:
                    new_qconfig_property["atomic_type"] = qgu.get_atomic_type(prop)
                elif property_type == qgu.QCONFIG_ENUMERATION_PROPERTY_TYPE:
                    # enumeration_target is used as a placeholder for enumeration member values
                    # it is referenced in "on_enum_parse" below and then removed
                    new_qconfig_property["enumeration_target"] = "{0}.{1}".format(
                        prop.type.name_of_package,
                        prop.type.name_of_type
                    )
                    new_qconfig_property["enumeration_is_open"] = True  # TODO: HOW ARE OPEN ENUMERATIONS ENCODED?
                elif property_type == qgu.QCONFIG_RELATIONSHIP_PROPERTY_TYPE:
                    new_qconfig_property["relationship_targets"] = []
                    for relationship_target_class in qgu.get_relationship_target_classes(prop):

                        new_qconfig_property["relationship_targets"].append("{0}.{1}".format(
                            relationship_target_class.package,
                            relationship_target_class.name
                        ))

                new_qconfig_class["properties"]["defined"].append(new_qconfig_property)

            qconfig_classes.append(new_qconfig_class)

    def on_enum_parse(self, ctx):
        """Event handler for the enum parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        enum = ctx.enum
        fully_qualified_enum_name = "{0}.{1}".format(enum.package.name, enum.name)
        qconfig = ctx.qconfig
        for qconfig_class in qconfig["classes"]["defined"]:
            for qconfig_property in qconfig_class["properties"]["defined"]:
                if qconfig_property["property_type"] == qgu.QCONFIG_ENUMERATION_PROPERTY_TYPE:
                    if "enumeration_target" in qconfig_property and qconfig_property["enumeration_target"] == fully_qualified_enum_name:
                        qconfig_property.pop("enumeration_target")
                        new_qconfig_enum_members = []
                        for i, member in enumerate(enum.members, start=1):
                            new_qconfig_enum_member = OrderedDict()
                            new_qconfig_enum_member["value"] = member.name
                            new_qconfig_enum_member["documentation"] = member.doc_string
                            new_qconfig_enum_member["order"] = i
                            new_qconfig_enum_members.append(new_qconfig_enum_member)
                        qconfig_property["enumeration_members"] = new_qconfig_enum_members



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.model.collection.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary collection, e.g. institute-id.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyesdoc
from pyesdoc.cv.model.entity import Entity



class Collection(Entity):
    """A vocabulary term collection.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.create_date = None
        self.description = None
        self.idx = None
        self.label = None
        self.name = None
        self.scope = None
        self.terms = list()
        self.uid = None
        self.url = None


    def __repr__(self):
        """Instance representation.

        """
        return self.namespace


    def __len__(self):
        """Returns number of terms in managed collection.

        """
        return len(self.terms)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return Entity.getiter(self.terms)


    def __getitem__(self, key):
        """Returns a child section item.

        """
        return Entity.getitem(self.terms, key)


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    @property
    def authority(self):
        """Gets associated governing authority.

        """
        return self.scope.authority


    @property
    def namespace(self):
        """Gets namespace.

        """
        return ":".join([
            self.scope.namespace,
            self.name,
            ])


    @property
    def full_idx(self):
        """Gets full computed idx.

        """
        return u"{}.{}".format(
            self.scope.idx,
            self.idx
            )


    @property
    def partition(self):
        """Returns associated partition.

        """
        return pyesdoc.cv.get_partition(self.namespace)



# -*- coding: utf-8 -*-

"""
.. module:: json.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: JSON codec.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.json.decoder import decode
from pyesdoc._codecs.json.encoder import encode



# -*- coding: utf-8 -*-

"""
.. module:: constants.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Supported ESDOC encodings.
ENCODING_DICT = 'dict'
ENCODING_JSON = 'json'
ENCODING_XML = 'xml'
ENCODING_HTML = 'html'
METAFOR_CIM_XML_ENCODING = 'xml-metafor-cim-v1'

# Standard ESDOC encodings.
ENCODINGS = (
    ENCODING_DICT,
    ENCODING_HTML,
    ENCODING_JSON,
    ENCODING_XML,
)

# Standard ESDOC file encodings.
ENCODINGS_FILE = (
    ENCODING_JSON,
    ENCODING_XML,
    ENCODING_HTML,
)

# Standard ESDOC http encodings.
ENCODINGS_HTTP = (
    ENCODING_JSON,
    ENCODING_XML,
    ENCODING_HTML,
)

# Custom ESDOC encodings.
ENCODINGS_CUSTOM = (
    METAFOR_CIM_XML_ENCODING,
)

# Custom ESDOC file encodings.
ENCODINGS_FILE_CUSTOM = (
    METAFOR_CIM_XML_ENCODING,
)

# All ESDOC encodings.
ENCODINGS_ALL = set(ENCODINGS + ENCODINGS_CUSTOM)

# Map of standard ESDOC encodings to MIME types.
ENCODING_HTTP_MEDIA_TYPES = {
    ENCODING_JSON : "application/json",
    ENCODING_XML : "application/xml",
    ENCODING_HTML : "text/html"
}

# Latest document version label.
DOC_VERSION_LATEST = 'latest'

# All document versions label.
DOC_VERSION_ALL = 'all'

# Default document encoding.
DEFAULT_ENCODING = 'json'

# Default document project.
DEFAULT_PROJECT = u'--'

# Default document institute.
DEFAULT_INSTITUTE = u'--'

# Default document source.
DEFAULT_SOURCE = u'scripts'

# Document viewer url.
VIEWER_URL_BY_ID = u"index.html?renderMethod=id&project={0}&id={1}&version={2}"

# Cardinality constraint.
CONSTRAINT_TYPE_CARDINALITY = 'cardinality'

# Constant constraint.
CONSTRAINT_TYPE_CONSTANT = 'constant'

# Type constraint.
CONSTRAINT_TYPE_TYPEOF = 'type'

# Regular expression constraint.
CONSTRAINT_TYPE_REGEX = 'regex'

# Set of supported constraint types.
CONSTRAINT_TYPES = {
    CONSTRAINT_TYPE_CARDINALITY,
    CONSTRAINT_TYPE_CONSTANT,
    CONSTRAINT_TYPE_REGEX,
    CONSTRAINT_TYPE_TYPEOF
}

# Cardinality constraint ... 0.0.
CARDINALITY_TYPE_0_0 = '0.0'

# Cardinality constraint ... 0.1.
CARDINALITY_TYPE_0_1 = '0.1'

# Cardinality constraint ... 1.1.
CARDINALITY_TYPE_1_1 = '1.1'

# Cardinality constraint ... 0.N.
CARDINALITY_TYPE_0_N = '0.N'

# Cardinality constraint ... 1.N.
CARDINALITY_TYPE_1_N = '1.N'

# Set of supported cardinality types.
CARDINALITY_TYPES = {
    CARDINALITY_TYPE_0_0,
    CARDINALITY_TYPE_0_1,
    CARDINALITY_TYPE_1_1,
    CARDINALITY_TYPE_0_N,
    CARDINALITY_TYPE_1_N
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.utils.factory
   :platform: Unix, Windows
   :synopsis: Encapsulates process of instantiating objects, i.e. generators, schemas and ontologies.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import re

from pyesdoc.mp.ontologies.core import Class
from pyesdoc.mp.ontologies.core import ClassConstraint
from pyesdoc.mp.ontologies.core import ClassPrintString
from pyesdoc.mp.ontologies.core import ComputedProperty
from pyesdoc.mp.ontologies.core import Decoding
from pyesdoc.mp.ontologies.core import Enum
from pyesdoc.mp.ontologies.core import EnumMember
from pyesdoc.mp.ontologies.core import Ontology
from pyesdoc.mp.ontologies.core import Package
from pyesdoc.mp.ontologies.core import Property



# Comutation element reg-ex.
_RE_COMPUTED_ELEMENT = '^[a-z]+$'


def _get_functions(modules):
    """Returns a collection of function pointers declared within a module.

    """
    try:
        iter(modules)
    except TypeError:
        modules = [modules]

    result = set()
    for module in modules:
        result.update({m[1] for m in inspect.getmembers(module) if inspect.isfunction(m[1])})

    return result


def _get_type_definitions(modules):
    """Returns set of type definitions instantiated from a set of modules.

    """
    def _get_definition(func):
        """Returns a type definition."""
        result = func()
        result['name'] = func.__name__
        result['doc'] = func.__doc__.strip()

        return result

    return [_get_definition(i) for i in _get_functions(modules)]


def _get_package_definitions(schema):
    """Returns set of package definitions.

    """
    def _get_definition(func):
        """Returns a package definition.

        """
        defn = {
            'name': func.__name__,
            'doc': func.__doc__.strip(),
            'types': func()
        }

        return defn

    return [_get_definition(f) for f in _get_functions(schema)]


def _get_class_properties(class_):
    """Returns class property definitions.

    """
    result = []
    doc_strings = class_.get('doc_strings', dict())
    for prop in class_.get('properties', []):
        if len(prop) == 4:
            name, type_name, cardinality, doc_string = prop
        elif len(prop) == 3:
            name, type_name, cardinality = prop
            doc_string = doc_strings.get(name, None)
        result.append(Property(name, type_name, cardinality, doc_string))
    if class_.get('is_document', False):
        result.append(Property('meta', 'shared.doc_meta_info', '1.1', "Injected document metadata."))

    return result


def _get_class_computed_properties(class_):
    """Returns class computed property definitions.

    """
    def get_computation(computation):
        """Returns parsed computation statement.

        """
        return " ".join(["self.{}".format(e) if re.match(_RE_COMPUTED_ELEMENT, e[0]) else e 
                        for e in computation.split(" ")])

    return [ComputedProperty(name, get_computation(computation)) for name, computation in
            class_.get('derived', []) + class_.get('computed', [])]


def _get_class_constraints(class_):
    """Returns class constraint definitions.

    """
    def get_constraint(defn):
        """Returns parsed computation statement.

        """
        property_name, typeof, value = defn

        return ClassConstraint(property_name, typeof, value)

    return [get_constraint(defn) for defn in class_.get('constraints', [])]


def _get_class_decodings(class_):
    """Returns class decoding definitions.

    """
    result = []
    for decoding in class_.get('decodings', []):
        result.append(Decoding(decoding[0],
                               decoding[1],
                               None if len(decoding) == 2 else decoding[2]))

    return result


def _get_class_pstr(class_):
    """Returns class decoding definitions.

    """
    pstr = class_.get('pstr')
    if pstr:
        return ClassPrintString(pstr[0], pstr[1])


def _get_package_classes(schema, package, types):
    """Returns package class definitions.

    """
    result = []
    for cls in [t for t in types if t['type'] == 'class']:
        result.append(
            Class(cls['name'],
                  cls.get('base', None),
                  cls.get('is_abstract', False),
                  cls.get('is_document', False),
                  cls.get('doc', None),
                  _get_class_properties(cls),
                  _get_class_computed_properties(cls),
                  _get_class_constraints(cls),
                  _get_class_decodings(cls),
                  _get_class_pstr(cls)
                  )
        )

    return result


def _get_package_enums(schema, package, types):
    """Returns package enum definitions.

    """
    result = []
    for enum in [t for t in types if t['type'] == 'enum']:
        result.append(
            Enum(enum['name'],
                 enum.get('is_open', True),
                 enum.get('doc', None),
                 [EnumMember(m[0], m[1]) for m in enum.get('members', [])])
            )

    return result


def _get_package_types(schema, package):
    """Returns package type definitions.

    """
    types = _get_type_definitions(package.get('types', []))

    return _get_package_classes(schema, package, types), \
           _get_package_enums(schema, package, types)


def _get_ontology_packages(schema):
    """Returns ontology package definitions.

    """
    result = []
    for package in _get_package_definitions(schema):
        classes, enums = _get_package_types(schema, package)
        result.append(
            Package(package['name'],
                    package['doc'],
                    classes,
                    enums
                    )
            )

    return result


def create_ontology(schema):
    """Factory method to instantiate an ontology instance from a schema declaration.

    :param module schema: An ontology schema declaration.

    :returns: An ontology declaration.
    :rtype: pyesdoc.mp.ontologies.core.Ontology

    """
    return Ontology(schema.NAME,
                    schema.VERSION,
                    schema.DOC,
                    _get_ontology_packages(schema))



# -*- coding: utf-8 -*-

"""
.. module:: file_info.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
    :synopsis: An archived document file wrapper.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import glob
import os

import pyesdoc
from pyesdoc import exceptions



class ArchiveFileInfo(object):
    """Represents an archived file.

    """
    def __init__(self, folder, path):
        """Instance constructor.

        """
        name = path.split("/")[-1]
        parts = name.split('.')[0].split("_")

        self.encoding = name.split('.')[-1]
        self.hashid = parts[3] if len(parts) == 4 else parts[0]
        self.folder = folder
        self.name = name
        self.parts = parts
        self.path = path
        self.path_error = "{}.ERROR".format(path.split('.')[0])
        self.type = parts[0] if len(parts) == 4 else None
        self.uid = parts[1] if len(parts) == 4 else None
        self.version = parts[2] if len(parts) == 4 else None


    def __repr__(self):
        """Instance representation.

        """
        return self.path


    @property
    def exists(self):
        """Gets flag indicating whether a matching file exists.

        """
        if os.path.exists(self.path):
            return True
        else:
            path = "*{0}.{1}".format(self.hashid, self.encoding)
            path = os.path.join(self.folder.path, path)
            return len(glob.glob(path)) == 1

    @property
    def is_raw(self):
        """Gets flag indicating whether the file is a raw unparsed file.

        """
        return len(self.parts) == 1

    @property
    def project(self):
        """Gets project with which file is associated.

        """
        return self.folder.project

    @property
    def source(self):
        """Gets source with which file is associated.

        """
        return self.folder.source


    def delete(self):
        """Deletes underlying file from local file system.

        """
        if self.exists:
            os.remove(self.path)


    def get_document(self, extend=True):
        """Returns deserialized document.

        :returns: Deserialized archived document.
        :rtype: object

        """
        # Read from file system.
        try:
            document = pyesdoc.read(self.path)
        except Exception as err:
            raise exceptions.LoadingException(err)

        # Parse - performs project/source specific processing.
        try:
            pyesdoc.archive.parse(document, self.encoding, self.project, self.source)
        except Exception as err:
            raise exceptions.ParsingException(err)

        # Extend - performs generic processing.
        if extend:
            try:
                pyesdoc.extend(document)
            except exceptions.ExtendingException as err:
                raise err
            except Exception as err:
                raise exceptions.ExtendingException(err)

        return document


    def get_content(self, encoding):
        """Returns content of an archived file in the desired encoding.

        :param str encoding: Desired document content encoding.

        :returns: The contents of an archived file in the desired encoding.
        :rtype: unicode

        """
        return pyesdoc.encode(self.get_document(), encoding)



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""

from collections import OrderedDict
DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()


# --------------------------------------------------------------------
# CONTACT
#
# Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS
#
# Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = ''

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS
#
# Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Top of atmosphere solar insolation characteristics'

# --------------------------------------------------------------------
# SUB-PROCESS: solar_forcing_pathways
# --------------------------------------------------------------------
DETAILS['solar_pathways'] = {
    'description': "Pathways for solar forcing of the atmosphere",
    'properties': [
        ('pathways', 'ENUM:solar_forcing_pathways', '1.N',
            'Pathways for the solar forcing of the atmosphere model domain'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: solar_constant
# --------------------------------------------------------------------
DETAILS['solar_constant'] = {
    'description': "Solar constant and top of atmosphere insolation characteristics",
    'properties': [
        ('type', 'ENUM:top_insolation_solar_constant_type', '1.1',
             'Time adaptation of the solar constant.'),
        ('fixed_value', 'float', '0.1',
             'If the solar constant is fixed, enter the value of the solar constant (W m-2).'),
        ('transient_characteristics', 'str', '1.1',
             'solar constant transient characteristics (W m-2)'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: orbital_parameters
# --------------------------------------------------------------------
DETAILS['orbital_parameters'] = {
    'description': "Orbital parameters and top of atmosphere insolation characteristics",
    'properties': [
        ('type', 'ENUM:top_insolation_orbital_parameters_type', '1.1',
            'Time adaptation of orbital parameters'),
        ('fixed_reference_date', 'int', '1.1',
            'Reference date for fixed orbital parameters (yyyy)'),
        ('solar_constant_transient_characteristics', 'str', '1.1',
            'Characteristics of transient orbital parameters'),
        ('computation_method', 'ENUM:top_insolation_orbital_parameters_computation_method', '1.1',
            'Method used for computing orbital parameters.')
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: insolation_ozone
# --------------------------------------------------------------------
DETAILS['insolation_ozone'] = {
    'description': "Impact of solar insolation on stratospheric ozone",
    'properties': [
        ('solar_ozone_impact', 'bool', '1.1',
            'Does top of atmosphere insolation impact on stratospheric ozone?'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------

ENUMERATIONS['solar_forcing_pathways'] = {
    'description': "Pathways for solar forcing of the atmosphere",
    'is_open': True,
    'members': [
        ('SW radiation', None),
        ('precipitating energetic particles', 'Precipitating energetic particles from the sun (predominantly protons) '
                            'and the magnetosphere (predominantly electrons) affect the ionization levels in the polar '
                            'middle and upper atmosphere, leading to significant changes of the chemical composition'),
        ('cosmic rays', 'Cosmic rays are the main source of ionization in the troposphere and lower stratosphere.')
    ]
}

ENUMERATIONS['top_insolation_solar_constant_type'] = {
    'description': "Time adaptation of the solar constant",
    'is_open': False,
    'members': [
        ('fixed', None),
        ('transient', None),
        ]
    }

ENUMERATIONS['top_insolation_orbital_parameters_type'] = {
    'description': "Time adaptation of orbital parameters",
    'is_open': False,
    'members': [
        ('fixed', None),
        ('transient', None),
        ]
    }

ENUMERATIONS['top_insolation_orbital_parameters_computation_method'] = {
    'description': "Method used for computing orbital parameters",
    'is_open': True,
    'members': [
        ('Berger 1978', None),
        ('Laskar 2004', None),
        ]
    }




# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_shared_platform.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.Platform instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.Platform

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Platform"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.shared.Platform.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = 'b765775a-e2ac-11df-9efb-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-04-23 14:59:07.049089').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, "IBM Power 6_Other")
    tu.assert_str(ext.description, "Machine IBM Power 6", True)
    tu.assert_str(ext.full_display_name, "CMIP5 Platform : MOHC - IBM Power 6_Other")
    tu.assert_int(ext.summary_fields, 2)
    tu.assert_str(ext.summary_fields[0], "IBM Power 6_Other")
    tu.assert_str(ext.summary_fields[1], "Machine IBM Power 6", True)


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_str(doc.long_name, "Machine IBM Power 6 and compiler Other")
    tu.assert_str(doc.short_name, "IBM Power 6_Other")
    tu.assert_iter(doc.contacts, 1, cim.v1.ResponsibleParty)
    tu.assert_str(doc.contacts[0].abbreviation, "MOHC")
    tu.assert_str(doc.contacts[0].organisation_name, "UK Met Office Hadley Centre")
    tu.assert_str(doc.contacts[0].role, "contact")
    tu.assert_iter(doc.units, 1, cim.v1.MachineCompilerUnit)


def _assert_doc_machine(doc, is_update):
    """Assert machine information."""
    tu.assert_object(doc.units[0].machine, cim.v1.Machine)
    m = doc.units[0].machine
    tu.assert_int(m.cores_per_processor, 32)
    tu.assert_str(m.interconnect, "Infiniband")
    tu.assert_str(m.name, "IBM Power 6")
    tu.assert_iter(m.libraries, 0)
    tu.assert_int(m.maximum_processors, 2)
    tu.assert_str(m.operating_system, "AIX")
    tu.assert_str(m.system, "Parallel")
    tu.assert_str(m.vendor, "IBM")
    tu.assert_str(m.processor_type, "Other")


def _assert_doc_compiler(doc, is_update):
    """Assert compiler information."""
    tu.assert_iter(doc.units[0].compilers, 1, cim.v1.Compiler)
    c = doc.units[0].compilers[0]
    tu.assert_str(c.name, "Other")
    tu.assert_str(c.version, "12.1.0.0")


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_machine,
        _assert_doc_compiler,
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


# -*- coding: utf-8 -*-
"""
.. module:: populate_from_feeds.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Populates archive with documents pulled from 3rd party ATOM feeds.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import hashlib
from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool

import feedparser

from pyesdoc.archive import manager as archive
from pyesdoc.archive import config
from pyesdoc.archive import feed_entry_processor
from pyesdoc.utils import runtime



class _FeedEntryInfo(object):
    """Encapsulates feed entry information.

    """
    def __init__(self, project, feed, entry):
        """Object constructor.

        """
        self.encoding = feed.encoding
        self.project = project.name
        self.source = feed.source
        self.url = entry.links[0]['href']
        self.url_hash = hashlib.md5(self.url).hexdigest()


class _DocumentProcessingInfo(object):
    """Encapsulates document processing information.

    """
    def __init__(self, archive, entry, index, verbose):
        """Object constructor.

        :param ArchiveInfo archive: Archive manager.
        :param _FeedEntryInfo entry: Feed entry information.
        :param int index: Document processing id.
        :param bool verbose: Logging level flag.

        """
        self.content = None
        self.download = None
        self.entry = entry
        self.error = None
        self.file = archive.get_file(entry.project,
                                     entry.source,
                                     entry.url_hash,
                                     entry.encoding)
        self.index = index
        self.project = entry.project
        self.source = entry.source
        self.verbose = verbose


    def __repr__(self):
        """Object representation.

        """
        return "{0} :: {1} :: {2} :: {3} :: {4}".format(
            self.project,
            self.source,
            self.file.name,
            self.entry.url,
            self.file.exists
            )

    @property
    def folder(self):
        """Gets associated archive folder.

        """
        return self.file.folder


def _log(msg):
    """Logging helper function.

    """
    runtime.log(msg)


def _yield_feeds(project_filter=None):
    """Yields feeds to be processed.

    """
    for project in config.get_projects():
        if project_filter and project.name != project_filter:
            continue
        for feed in [f for f in project.feeds if f.is_active]:
            msg = "processing feed: {0} --> {1} --> {2}"
            msg = msg.format(project.name, feed.source, feed.url)
            _log(msg)
            yield project, feed


def _yield_feed_entries(project_filter=None):
    """Yields active feed entries.

    """
    for project, feed in _yield_feeds(project_filter):
        for entry in feedparser.parse(feed.url).entries:
            yield _FeedEntryInfo(project, feed, entry)


def _yield_documents(archive, project, throttle, verbose):
    """Yields documents for processing.

    """
    scanned = yielded = 0
    for entry in _yield_feed_entries(project):
        scanned += 1
        ctx = _DocumentProcessingInfo(archive, entry, yielded + 1, verbose)
        if not ctx.file.exists:
            yielded += 1
            yield ctx

        # ... apply throttle
        if throttle and throttle == yielded:
            break

    # Log processing stats.
    msg = "{0} feed entries scanned: {1} requires processing"
    msg = msg.format(scanned, yielded)
    _log(msg)
    errors = archive.get_error_count()
    if errors:
        msg = "WARNING :: {0} processing errors occurred".format(errors)
        _log(msg)


def execute(throttle, project=None, in_parallel=True):
    """Writes pulled documents to local file system.

    :param int throttle: A limit upon the number of documents to pull.
    :param str project: Name of a supported project (e.g. cmip6).
    :param bool in_parallel: Flag indicating whether job will leverage multiple cores.

    """
    archive.delete_error_files()
    pool = Pool() if in_parallel else ThreadPool()
    pool.map(feed_entry_processor.process,
             _yield_documents(archive, project, throttle, throttle > 0))
    pool.close()



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.utils.convert.py
   :copyright: Copyright "Feb 7, 2013", Earth System Documentation
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of library conversion functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import datetime
import json
import re
import types
import uuid

import arrow



# Default character set.
_JSON_CHARSET = "ISO-8859-1"

# Set of data types to be ignored when encoding.
_ENCODE_IGNOREABLE = (int, float, long, type(None), unicode)

# Set of data types to be converted to unicode when encoding.
_ENCODE_UNICODEABLE = (
    basestring,
    arrow.Arrow,
    datetime.datetime,
    uuid.UUID
    )

# ISO date formats.
_ISO_DATE_FORMATS = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S"]

# Values considered to be abbreviations.
_ABBREVIATIONS = ("id", "uid", "uuid")

# Default string encoding.
_UNICODE = "utf-8"


def str_to_unicode(val):
    """Converts string input to a unicode literal.

    :param object val: arget to be converted to a unicode literal.

    :returns: A unicode literal.
    :rtype: unicode

    """
    if val is None:
        return unicode()
    if isinstance(val, unicode):
        return val

    val = str(val).decode(_UNICODE).strip()
    if not len(val):
        return unicode()

    return unicode(val)


def str_to_camel_case(target, separator='_'):
    """Converts passed name to camel case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to camel case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Convert to pascal case.
        s = str_to_pascal_case(target, separator)

        # Preserve initial separator
        if s[0:len(separator)] == separator:
            r += separator
            s = s[len(separator):]

        # Lower case abbreviations.
        if s.lower() in _ABBREVIATIONS:
            r += s.lower()

        # Lower case initial character.
        elif len(s):
            r += s[0].lower()
            r += s[1:]

    return r


def str_to_underscore_case(target):
    """Helper function to convert a from camel case string to an underscore case string.

    :param target: A string for conversion.
    :type target: str

    :returns: A string converted to underscore case, e.g. account_number.
    :rtype: str

    """
    if target is None or not len(target):
        return ''

    r = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', target)
    r = re.sub('([a-z0-9])([A-Z])', r'\1_\2', r)
    r = r.lower()

    return r


def str_to_pascal_case(target, separator='_'):
    """Converts passed name to pascal case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to pascal case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Preserve initial separator
        if target[0:len(separator)] == separator:
            r = separator

        # Iterate string parts.
        s = target.split(separator)
        for s in s:

            # Upper case abbreviations.
            if s.lower() in _ABBREVIATIONS:
                r += s.upper()

            # Upper case initial character.
            elif (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]

    return r


def _to_encodable(obj, key_formatter=str_to_camel_case):
    """Converts data to encodeable representation.

    """
    if isinstance(obj, _ENCODE_IGNOREABLE):
        return obj

    elif isinstance(obj, _ENCODE_UNICODEABLE):
        return unicode(obj)

    elif isinstance(obj, collections.Mapping):
        return {unicode(key_formatter(k)): _to_encodable(v) for k, v in iter(obj.items())}

    elif isinstance(obj, collections.Iterable):
        return [_to_encodable(i) for i in obj]


class _JSONDecoder(json.JSONDecoder):
    """Extends json decoder so as to handle extended types.

    """
    def __init__(self, key_formatter, to_namedtuple=False):
        """Instance constructor.

        """
        json.JSONDecoder.__init__(self,
                                  encoding=_JSON_CHARSET,
                                  object_hook=self.dict_to_object)
        self.key_formatter = key_formatter
        self.to_namedtuple = to_namedtuple
        self.value_parsers = [
            self.unicode_to_datetime,
            self.unicode_to_uuid
            ]


    def dict_to_object(self, d):
        """Converts a dictionary to an object.

        """
        # Parse values.
        for k, v in d.items():
            for parser in self.value_parsers:
                if parser(d, k, v):
                    break

        # Format keys.
        if self.key_formatter is not None:
            d = dict_keys(d, self.key_formatter)

        # Return dictionary | named tuple.
        return d


    def unicode_to_datetime(self, d, k, v):
        """Converts a unicode value to datetime.

        """
        if isinstance(v, unicode) and len(v):
            try:
                float(v)
            except ValueError:
                for format in _ISO_DATE_FORMATS:
                    try:
                        v = datetime.datetime.strptime(v, format)
                    except (ValueError, TypeError):
                        pass
                    else:
                        d[k] = v
                        return True

        return False


    def unicode_to_uuid(self, d, k, v):
        """Converts a unicode value to uuid.UUID.

        """
        if isinstance(v, unicode) and len(v):
            try:
                v = uuid.UUID(v)
            except ValueError:
                pass
            else:
                d[k] = v
                return True

        return False


def json_to_dict(as_json, key_formatter=None):
    """Converts a json encoded string to a dictionary.

    :param unicode as_json: A json encoded string.
    :param function key_formatter: Dictionary key formatter.

    :returns: A dictionary.
    :rtype: dict

    """
    return _JSONDecoder(key_formatter).decode(as_json)


def dict_to_json(as_dict):
    """Converts a dictionary to json.

    :param dict as_dict: A dictionary.

    :returns: A json encoded text blob.
    :rtype: unicode

    """
    return json.dumps(_to_encodable(as_dict), indent=4, sort_keys=True)


def dict_keys(as_dict, key_formatter=str_to_pascal_case):
    """Converts keys of a dictionary using the passed key formatter.

    :param dict as_dict: A dictionary.
    :param function key_formatter: A dictionary key formatter function pointer.

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    if not isinstance(as_dict, dict):
        return as_dict

    result = {}
    for key, val in as_dict.items():
        if isinstance(val, collections.Mapping):
            result[key_formatter(key)] = dict_keys(val, key_formatter)
        elif isinstance(val, types.ListType):
            result[key_formatter(key)] = [dict_keys(i, key_formatter) for i in val]
        else:
            result[key_formatter(key)] = val

    return result



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.generator
   :platform: Unix, Windows
   :synopsis: Base class encapsulating functionality common to all cim code generators.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from abc import ABCMeta

from pyesdoc.mp.ontologies.generators import generator_utils as gu



class Generator(object):
    """Base class encapsulating functionality common to all code generators.

    """
    # Abstract Base Class module - see http://docs.python.org/library/abc.html
    __metaclass__ = ABCMeta

    def execute(self, ctx):
        """Executes the code generator.

        :param pyesdoc.mp.ontologies.core.Ontology ontology: Ontology being processed.

        """
        # Escape if not required.
        if not self.is_required(ctx):
            return

        # Notify start.
        self.on_start(ctx)

        # Raise parsing events and emit code accordingly.
        self.on_ontology_parse(ctx)
        for pkg in ctx.ontology.packages:
            ctx.set_package(pkg)
            self.on_package_parse(ctx)
        for cls in ctx.ontology.classes:
            ctx.set_class(cls)
            self.on_class_parse(ctx)
        for enum in ctx.ontology.enums:
            ctx.set_enum(enum)
            self.on_enum_parse(ctx)

        # Notify end.
        self.on_end(ctx)


    def _write_code(self, ctx, code):
        """Writes generated code to file system.

        """
        if code:
            for code, dir_, fpath in code:
                gu.write_file(gu.format_code(ctx, code), dir_, fpath)


    def is_required(self, ctx):
        """Predicate determing whether code generation is actually required.

        """
        return True


    def on_start(self, ctx):
        """Event handler for the parsing start event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass


    def on_end(self, ctx):
        """Event handler for the parsing end event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass


    def on_ontology_parse(self, ctx):
        """Event handler for the ontology parse event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass


    def on_package_parse(self, ctx):
        """Event handler for the package parse event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass


    def on_class_parse(self, ctx):
        """Event handler for the class parse event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass


    def on_enum_parse(self, ctx):
        """Event handler for the enum parse event.

        :param pyesdoc.mp.ontologies.generators.generator.GeneratorContext ctx: Generation context information.

        """
        pass



# -*- coding: utf-8 -*-

"""
.. module:: test_general.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc.cv general tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import uuid

import nose.tools

import pyesdoc.cv as LIB
import tests.cv.utils as tu



# Set of classes exposed by library.
_CLASSES = {
    'Term'
}

# Set of constants exposed by library.
_CONSTANTS = {
    'GOVERNANCE_STATUS_ACCEPTED',
    'GOVERNANCE_STATUS_DEPRECATED',
    'GOVERNANCE_STATUS_PENDING',
    'GOVERNANCE_STATUS_REJECTED',
    'OPT_IO_DIR',
    'OPT_IS_VERBOSE'
}

# Set of exceptions exposed by library.
_EXCEPTIONS = {
    'InvalidAssociationError',
    'InvalidOptionError',
    'ValidationError'
}

# Set of functions exposed by library.
_FUNCS = {
    'accept',
    'create',
    'delete',
    'deprecate',
    'get_count',
    'get_option',
    'get_term',
    'get_termset',
    'get_termsets',
    'init',
    'is_valid',
    'reject',
    'reset',
    'save',
    'validate'
}


def test_library_exports():
    """Test set of exports exposed by library.

    """
    def _test_member(member, member_type):
        """Test that library exposes the named member."""
        assertor = getattr(tu, 'assert_has_{}'.format(member_type))
        assertor(LIB, member)

    for members, member_type, in (
        (_CLASSES, 'class'),
        (_CONSTANTS, 'constant'),
        (_EXCEPTIONS, 'exception'),
        (_FUNCS, 'function'),
        ):
        for member in members:
            desc = "exposes a {} called {}".format(member_type, member)
            tu.init(_test_member, 'library', desc)
            yield _test_member, member, member_type


def test_library_initialization():
    """Test library initialization.

    """
    def _test_init_01():
        """Normal initialization."""
        with tu.get_options() as opts:
            LIB.init(opts)

    @nose.tools.raises(IOError)
    def _test_init_02():
        """Initialization error caused by invalid io directory."""
        LIB.init({
            'io_dir': uuid.uuid4()
        })

    @nose.tools.raises(ValueError)
    def _test_init_03():
        """Initialization error caused by invalid verbose flag."""
        LIB.init({
            'verbose': float()
        })

    for func in (_test_init_01, _test_init_02, _test_init_03):
        tu.init(func, 'library')
        yield func



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._factory.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document creation functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime
import uuid

from pyesdoc import exceptions
from pyesdoc import ontologies



def create(
    typeof,
    project=None,
    sub_projects=None,
    institute=None,
    source=None,
    author=None,
    uid=None,
    version=0
    ):
    """Creates a document.

    :param class typeof: Ontology type, e.g. cim.1.software.ModelComponent.
    :param str project: Project wih which instance is associated.
    :param list|str sub_projects: Sub-project(s) with which instance is associated.
    :param str institute: Institute wih which instance is associated.
    :param str source: Source application with which instance is associated.
    :param str author: Author wih which instance is associated.
    :param uuid.UUID uid: Document unique identifier.

    :returns: A pyesdoc document instance.
    :rtype: pyesdoc object

    """
    # Validate document type.
    if typeof not in ontologies.type_info.TYPES:
        raise exceptions.InvalidDocumentTypeException(typeof)

    # Set defaults.
    if not uid:
        uid = uuid.uuid4()

    # Reformat.
    if institute:
        institute = unicode(institute).lower()
    if project:
        project = unicode(project).lower()
    if source:
        source = unicode(source).lower()
    elif institute:
        source = institute
    elif project:
        source = project
    uid = unicode(uid)
    if sub_projects:
        if isinstance(sub_projects, (str, unicode)):
            sub_projects = [sub_projects]
        sub_projects = [unicode(i).lower() for i in sub_projects]

    # Verify uid.
    try:
        uuid.UUID(uid)
    except ValueError:
        raise ValueError("Document identifiers must be UUID's.")

    # Instantiate & initialize meta info (if necessary).
    doc = typeof()
    if hasattr(doc, 'meta'):
        doc.meta.author = author
        doc.meta.id = uid
        doc.meta.version = version
        doc.meta.create_date = datetime.datetime.now()
        doc.meta.institute = institute or None
        doc.meta.project = project or None
        doc.meta.sub_projects = sub_projects or []
        doc.meta.source = source or None
        doc.meta.type = typeof.type_key

    return doc



# -*- coding: utf-8 -*-

"""
.. module:: test_authoring.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv authoring tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect

import nose
import pyessv

import tests.utils as tu


def _setup():
    """Test runner setup.

    """
    tu.setup()
    count = pyessv.get_count()
    term = tu.create_term()
    pyessv.save(term)
    tu.assert_int(pyessv.get_count(), count + 1)


def test_create():
    """Test creating a term.

    """
    @nose.with_setup(tu.setup, tu.teardown)
    def test_create_01():
        """Create term."""
        term = tu.create_term()
        tu.assert_object(term, pyessv.Term)
        tu.assert_int(pyessv.get_count(), 0)


    @nose.with_setup(_setup, tu.teardown)
    def test_create_02():
        """Create & save term."""
        term = pyessv.get_term(tu.TERM_DOMAIN, tu.TERM_SUBDOMAIN, tu.TERM_KIND, tu.TERM_NAME)
        tu.assert_object(term, pyessv.Term)

    for test in (test_create_01, test_create_02):
        tu.init(test, 'authoring', inspect.getdoc(test))
        yield test


@nose.with_setup(_setup, tu.teardown)
def test_delete():
    """pyessv-tests: authoring: deleting a term.

    """
    count = pyessv.get_count()
    term = tu.get_term()
    pyessv.delete(term)
    tu.assert_int(pyessv.get_count(), count - 1)
    term = tu.get_term()
    tu.assert_none(term)


def test_update():
    """Test updating a term.

    """
    def _update_create_date(term):
        """Update term create date."""
        term.create_date = tu.get_date()

    def _update_description(term):
        """Update term description."""
        term.description = tu.get_uuid()

    def _update_domain(term):
        """Update term domain."""
        term.domain = tu.get_unicode(existing=term.domain)

    def _update_id(term):
        """Update term id."""
        term.idx = tu.get_int(existing=term.idx)

    def _update_kind(term):
        """Update term kind."""
        term.kind = tu.get_unicode(existing=term.kind)

    def _update_name(term):
        """Update term name."""
        term.name = tu.get_unicode(existing=term.name)

    def _update_status(term):
        """Update term status."""
        term.status = tu.get_unicode(existing=term.status)

    def _update_subdomain(term):
        """Update term subdomain."""
        term.subdomain = tu.get_unicode(existing=term.subdomain)

    def _update_uid(term):
        """Update term uid."""
        term.uid = tu.get_uuid()

    def _assert(term):
        """Asserts an update."""
        count = pyessv.get_count()
        pyessv.save(term)
        tu.assert_int(pyessv.get_count(), count)
        term_ = pyessv.get_term(term.domain, term.subdomain, term.kind, term.name)
        tu.assert_terms(term, term_)

    @nose.with_setup(_setup, tu.teardown)
    def _test(update_callback):
        """Performs update test."""
        term = tu.get_term()
        update_callback(term)
        _assert(term)

    for func  in (
        _update_create_date,
        _update_description,
        _update_domain,
        _update_id,
        _update_kind,
        _update_name,
        _update_status,
        _update_subdomain,
        _update_uid
        ):
        tu.init(_test, 'authoring', inspect.getdoc(func))
        yield _test, func



# -*- coding: utf-8 -*-

"""
.. module:: software_classes.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 software package class definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""


def component_language():
    """Details of the programming language a component is written in. There is an assumption that all EntryPoints use the same ComponentLanguage.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('name', 'str', '1.1'),
            ('properties', 'software.component_language_property', '0.N'),
        ],
        'doc_strings': {
            'name': 'The name of the language.'
        },
        'decodings' : [
            ('name', 'child::cim:name'),
        ]
    }


def component_language_property():
    """This provides a place to include language-specific information. Every property is basically a name/value pair, where the names are things like: moduleName, reservedUnits, reservedNames (these are all examples of Fortran-specific properties).

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False
    }


def component_property():
    """ComponentProperties include things that a component simulates (ie: pressure, humidity) and things that prescribe that simulation (ie: gravity, choice of advection scheme). Note that this is a specialisation of shared::DataSource. data::DataObject is also a specialisation of shared::DataSource. This allows software::Connections and/or activity::Conformance to refer to either ComponentProperties or DataObjects.

    """
    return {
        'type' : 'class',
        'base' : 'shared.data_source',
        'is_abstract' : False,
        'properties' : [
            ('sub_properties', 'software.component_property', '0.N'),
            ('citations', 'shared.citation', '0.N'),
            ('description', 'str', '0.1'),
            ('grid', 'str', '0.1'),
            ('intent', 'software.component_property_intent_type', '0.1'),
            ('is_represented', 'bool', '1.1'),
            ('long_name', 'str', '0.1'),
            ('short_name', 'str', '1.1'),
            ('standard_names', 'str', '0.N'),
            ('units', 'shared.unit_type', '0.1'),
            ('values', 'str', '0.N'),
        ],
        'doc_strings': {
            'grid': 'A reference to the grid that is used by this component.',
            'intent': 'The direction that this property is intended to be coupled: in, out, or inout.',
            'is_represented': 'When set to false, means that this property is not used by the component. Covers the case when, for instance, a modeler chooses not to represent some property in their model. (But still allows meaningful comparisons between components which _do_ model this property.)',
            'units': 'The standard name that this property is known as (for example, its CF name).',
            'values': 'The value of the property (not applicable to fields).',
        },
        'decodings' : [
            ('sub_properties', 'child::cim:componentProperty'),
            ('citations', 'child::cim:citation'),
            ('description', 'child::cim:description'),
            ('grid'),
            ('intent', 'self::cim:componentProperty/@intent'),
            ('is_represented', 'self::cim:componentProperty/@represented'),
            ('long_name', 'child::cim:longName'),
            ('short_name', 'child::cim:shortName'),
            ('standard_names', 'child::cim:standardName/@value'),
            ('units', 'child::cim:units/@value'),
            ('values', 'child::cim:value'),
        ]
    }


def composition():
    """The set of Couplings used by a Component. Couplings can only occur between child components. That is, a composition must belong to an ancestor component of the components whose fields are being connected.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('couplings', 'str', '0.N'),
            ('description', 'str', '0.1'),
        ]
    }


def connection():
    """A Connection represents a link from a source DataSource to a target DataSource.  These can either be ComponentProperties (ie: the values come from an internal component) or DataObjects (ie: the values come from an external file).   It can be associated with another software component (a transformer).  If present, the rate, lag, timeTransformation, and spatialRegridding override that of the parent coupling.  Note that there is the potential for multiple connectionSource & connectionTarget and multiple couplingSources & couplingTargets.  This may lead users to wonder how to match up a connection source (a ComponentProperty) with its coupling source (a SoftwareComponent). Clever logic is not required though; because the sources and targets are listed by reference, they can be found in a CIM document and the parent can be navigated to from there - there is no need to consult the source or target of the coupling.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('properties', 'software.connection_property', '0.N'),
            ('sources', 'software.connection_endpoint', '0.N'),
            ('target', 'software.connection_endpoint', '0.1'),
            ('description', 'str', '0.1'),
            ('priming', 'shared.data_source', '0.1'),
            ('spatial_regridding', 'software.spatial_regridding', '0.N'),
            ('time_lag', 'str', '0.1'),
            ('time_profile', 'software.timing', '0.1'),
            ('time_transformation', 'software.time_transformation', '0.1'),
            ('transformers', 'software.processor_component', '0.N'),
            ('type', 'software.connection_type', '0.1'),
        ],
        'doc_strings': {
            'sources': 'The source property being connected.  (note that there can be multiple sources)  This is optional; the file/component source may have already been specified by the couplingSource.',
            'target': 'The target property being connected.  This is optional to support the way that input is handled in the CMIP5 questionnaire.',
            'priming': 'A priming source is one that is active on the first available timestep only (before "proper" coupling can ocurr). It can either be described here explicitly, or else a separate coupling/connection with a timing profile that is active on only the first timestep can be created.',
            'spatial_regridding': 'Characteristics of the scheme used to interpolate a field from one grid (source grid) to another (target grid)',
            'time_lag': 'The coupling field used in the target at a given time corresponds to a field produced by the source at a previous time.',
            'time_profile': 'All information having to do with the rate of this connection; the times that it is active.  This overrides any rate of a Coupling.',
            'time_transformation': 'Temporal transformation performed on the coupling field before or after regridding onto the target grid.',
            'transformers': 'An "in-line" transformer. This references a fully-described transformer (typically that forms part of the top-level composition) used in the context of this coupling. It is used instead of separately specifying a spatialRegridding, timeTransformation, etc. here.',
            'type': 'The type of Connection'
        },
        'decodings' : [
            ('properties', 'child::cim:connectionProperty'),
            ('sources', 'child::cim:connectionSource'),
            ('target', 'child::cim:connectionTarget'),
            ('description', 'child::cim:description'),
            ('priming', 'child::cim:priming/cim:priming/cim:dataObject', 'data.data_object'),
            ('priming', 'child::cim:priming/cim:priming/cim:dataContent', 'data.data_content'),
            ('priming', 'child::cim:priming/cim:priming/cim:componentProperty', 'software.component_property'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.model_component'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.processor_component'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.statistical_model_component'),
            ('priming', 'child::cim:priming/cim:reference', 'shared.doc_reference'),
            ('spatial_regridding', 'child::cim:spatialRegridding'),
            ('time_lag', 'child::cim:timeLag'),
            ('time_profile', 'child::cim:timeProfile'),
            ('time_transformation', 'child::cim:timeTransformation'),
            ('transformers', 'child::cim:transformer/cim:processorComponent'),
            ('transformers', 'child::cim:transformer/cim:reference', 'shared.doc_reference'),
            ('type', 'child::cim:type/@value'),
        ]
    }


def connection_endpoint():
    """The source/target of a coupling.  This is a DataSource (a SoftwareComponent or DataObject).  This is a separate class in order to associate an instanceID with the DataSource; this is used to identify which particular instance is being coupled in case the same DataSource is used more than once in a coupled model (this may be required for BFG).

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('properties', 'software.connection_property', '0.N'),
            ('data_source', 'shared.data_source', '0.1'),
            ('instance_id', 'str', '0.1'),
        ],
        'doc_strings': {
            'properties': 'The place to describe features specific to the source/target of a connection.',
            'instance_id': 'If the same datasource is used more than once in a coupled model then a method for identifying which particular instance is being referenced is needed (for BFG).',
        },
        'decodings' : [
            ('properties', 'child::cim:connectionProperty'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:dataObject', 'data.data_object'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:dataContent', 'data.data_content'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:componentProperty', 'software.component_property'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.model_component'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.processor_component'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.statistical_model_component'),
            ('data_source', 'child::cim:dataSource/cim:reference', 'shared.doc_reference'),
            ('instance_id', 'child::cim:instanceID'),
        ]
    }


def connection_property():
    """A ConnectionProperty is a name/value pair used to specify OASIS-specific properties.

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False
    }


def coupling():
    """A coupling represents a set of Connections between a source and target component. Couplings can be complete or incomplete. If they are complete then they must include all Connections between model properties. If they are incomplete then the connections can be underspecified or not listed at all.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('connections', 'software.connection', '0.N'),
            ('properties', 'software.coupling_property', '0.N'),
            ('sources', 'software.coupling_endpoint', '1.N'),
            ('target', 'software.coupling_endpoint', '1.1'),
            ('description', 'str', '0.1'),
            ('is_fully_specified', 'bool', '1.1'),
            ('purpose', 'shared.data_purpose', '1.1'),
            ('priming', 'shared.data_source', '0.1'),
            ('spatial_regriddings', 'software.spatial_regridding', '0.N'),
            ('time_profile', 'software.timing', '0.1'),
            ('time_lag', 'software.time_lag', '0.1'),
            ('time_transformation', 'software.time_transformation', '0.1'),
            ('transformers', 'software.processor_component', '0.N'),
            ('type', 'software.connection_type', '0.1')
        ],
        'doc_strings': {
            'sources': 'The source component of the coupling. (note that there can be multiple sources).',
            'target': 'The target component of the coupling.',
            'description': 'A free-text description of the coupling.',
            'is_fully_specified': 'If true then the coupling is fully-specified.  If false then not every Connection has been described within the coupling.',
            'priming': 'A priming source is one that is active on the first available timestep only (before "proper" coupling can ocurr). It can either be described here explicitly, or else a separate coupling/connection with a timing profile that is active on only the first timestep can be created.',
            'spatial_regriddings': 'Characteristics of the scheme used to interpolate a field from one grid (source grid) to another (target grid).',
            'time_profile': 'Describes how often the coupling takes place.',
            'time_lag': 'The coupling field used in the target at a given time corresponds to a field produced by the source at a previous time.',
            'time_transformation': 'Temporal transformation performed on the coupling field before or after regridding onto the target grid.',
            'transformers': 'An "in-line" transformer. This references a fully-described transformer (typically that forms part of the top-level composition) used in the context of this coupling. It is used instead of separately specifying a spatialRegridding, timeTransformation, etc. here.',
            'type': 'Describes the method of coupling.'
        },
        'decodings' : [
            ('connections', 'child::cim:connection'),
            ('properties', 'child::cim:couplingProperty'),
            ('sources', 'child::cim:couplingSource'),
            ('target', 'child::cim:couplingTarget'),
            ('description', 'child::cim:description'),
            ('is_fully_specified', '@fullySpecified'),
            ('priming', 'child::cim:priming/cim:priming/cim:dataObject', 'data.data_object'),
            ('priming', 'child::cim:priming/cim:priming/cim:dataContent', 'data.data_content'),
            ('priming', 'child::cim:priming/cim:priming/cim:componentProperty', 'software.component_property'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.model_component'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.processor_component'),
            ('priming', 'child::cim:priming/cim:priming/cim:softwareComponent', 'software.statistical_model_component'),
            ('priming', 'child::cim:priming/cim:reference', 'shared.doc_reference'),
            ('purpose', '@purpose'),
            ('spatial_regriddings', 'child::cim:spatialRegridding'),
            ('time_lag', 'child::cim:timeLag'),
            ('time_profile', 'child::cim:timeProfile'),
            ('time_transformation', 'child::cim:timeTransformation'),
            ('transformers', 'child::cim:transformer/cim:processorComponent'),
            ('transformers', 'child::cim:transformer/cim:reference', 'shared.doc_reference'),
            ('type', 'child::cim:type/@value')
        ]
    }


def coupling_endpoint():
    """The source/target of a coupling.  This is a DataSource (a SoftwareComponent or DataObject).  This is a separate class in order to associate an instanceID with the DataSource; this is used to identify which particular instance is being coupled in case the same DataSource is used more than once in a coupled model (this may be required for BFG).

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('properties', 'software.coupling_property', '0.N'),
            ('data_source', 'shared.data_source', '0.1'),
            ('instance_id', 'str', '0.1')
        ],
        'doc_strings': {
            'properties': 'A place to describe features specific to the source/target of a coupling',
            'instance_id': 'If the same datasource is used more than once in a coupled model then a method for identifying which particular instance is being referenced is needed (for BFG).'
        },
        'decodings' : [
            ('properties', 'child::cim:couplingProperty'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:dataObject', 'data.data_object'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:dataContent', 'data.data_content'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:componentProperty', 'software.component_property'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.model_component'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.processor_component'),
            ('data_source', 'child::cim:dataSource/cim:dataSource/cim:softwareComponent', 'software.statistical_model_component'),
            ('data_source', 'child::cim:dataSource/cim:reference', 'shared.doc_reference'),
            ('instance_id', 'child::cim:instanceID')
        ]
    }


def coupling_property():
    """A CouplingProperty is a name/value pair used to specify OASIS-specific properties.

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False
    }


def deployment():
    """Gives information about the technical properties of a component: what machine it was run on, which compilers were used, how it was parallised, etc. A deployment basically associates a deploymentDate with a Platform. A deployment only exists if something has been deployed. A platform, in contrast, can exist independently, waiting to be used in deployments.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('deployment_date', 'datetime', '0.1'),
            ('description', 'str', '0.1'),
            ('parallelisation', 'software.parallelisation', '0.1'),
            ('platform', 'shared.platform', '0.1'),
            ('executable_name', 'str', '0.1'),
            ('executable_arguments', 'str', '0.N'),
        ],
        'doc_strings': {
            'platform': 'The platform that this deployment has been run on. It is optional to allow for "unconfigured" models, that nonetheless specify their parallelisation constraints (a feature needed by OASIS).'
        },
        'decodings' : [
            ('deployment_date', 'child::cim:deploymentDate'),
            ('description', 'child::cim:description'),
            ('parallelisation', 'child::cim:parallelisation'),
            ('platform', 'child::cim:platform/cim:platform'),
            ('platform', 'child::cim:platform/cim:reference', 'shared.doc_reference'),
            ('executable_name', 'child::cim:executableName'),
            ('executable_arguments', 'child::cim:executableArgument'),
        ]
    }


def entry_point():
    """Describes a function or subroutine of a SoftwareComponent. BFG will use these EntryPoints to define a schedule of subroutine calls for a coupled model. Currently, a very basic schedule can be approximated by using the "proceeds" and "follows" attributes, however a more complete system is required for full BFG compatibility. Every EntryPoint can have a set of arguments associated with it. These reference (previously defined) ComponentProperties.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('name', 'str', '0.1'),
        ]
    }


def model_component():
    """A ModelComponent is a scientific model; it represents code which models some physical phenomena for a particular length of time.

    """
    return {
        'type' : 'class',
        'base' : 'software.component',
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('type', 'software.model_component_type', '0.1'),
            ('types', 'software.model_component_type', '1.N'),
            ('timing', 'software.timing', '0.1'),
            ('activity', 'activity.activity', '0.1'),
        ],
        'doc_strings': {
            'type': 'Describes the type of component. There can be multiple types.',
            'types': 'Describes the type of component. There can be multiple types.',
            'timing': 'Describes information about how this component simulates time.',
        },
        'decodings' : [
            ('meta', 'self::cim:modelComponent'),
            ('timing', 'child::cim:timing'),
            ('type', 'child::cim:type[1]/@value'),
            ('types', 'child::cim:type/@value'),
        ]
    }


def parallelisation():
    """Describes how a deployment has been parallelised across a computing platform.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('processes', 'int', '1.1'),
            ('ranks', 'software.rank', '0.N'),
        ],
        'decodings' : [
            ('processes', 'child::cim:processes'),
            ('ranks', 'child::cim:rank'),
        ]
    }


def processor_component():
    """A ModelComponent is a scientific model; it represents code which models some physical phenomena for a particular length of time.

    """
    return {
        'type' : 'class',
        'base' : 'software.component',
        'is_abstract' : False,
        'is_document': True,
        'properties' : [

        ],
        'decodings' : [
            ('meta', 'self::cim:modelComponent'),
        ]
    }


def rank():
    """Creates and returns instance of rank class.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('value', 'int', '0.1'),
            ('min', 'int', '0.1'),
            ('max', 'int', '0.1'),
            ('increment', 'int', '0.1'),
        ],
        'decodings' : [
            ('value', 'child::cim:rankValue'),
            ('min', 'child::cim:rankMin'),
            ('max', 'child::cim:rankMax'),
            ('increment', 'child::cim:rankIncrement'),
        ]
    }


def component():
    """A SofwareComponent is an abstract component from which all other components derive. It represents an element that takes input data and generates output data. A SoftwareCompnent can include nested "child" components. Every component can have "componentProperties" which describe the scientific properties that a component simulates (for example, temperature, pressure, etc.) and the numerical properties that influence how a component performs its simulation (for example, the force of gravity). A SoftwareComponent can also have a Deployment, which describes how software is deployed onto computing resources. And a SoftwareComponent can have a composition, which describes how ComponentProperties are coupled together either to/from other SoftwareComponents or external data files. The properties specified by a component\'s composition must be owned by that component or a child of that component; child components cannot couple together their parents\' properties.

    """
    return {
        'type' : 'class',
        'base' : 'shared.data_source',
        'is_abstract' : False,
        'properties' : [
            ('sub_components', 'software.component', '0.N'),
            ('citations', 'shared.citation', '0.N'),
            ('code_access', 'str', '0.1'),
            ('composition', 'software.composition', '0.1'),
            ('coupling_framework', 'software.coupling_framework_type', '0.1'),
            ('dependencies', 'software.entry_point', '0.N'),
            ('deployments', 'software.deployment', '0.N'),
            ('description', 'str', '0.1'),
            ('funding_sources', 'str', '0.N'),
            ('grid', 'grids.grid_spec', '0.1'),
            ('is_embedded', 'bool', '0.1'),
            ('language', 'software.component_language', '0.1'),
            ('license', 'shared.license', '0.1'),
            ('long_name', 'str', '0.1'),
            ('online_resource', 'uri', '0.1'),
            ('parent', 'software.component', '0.1'),
            ('previous_version', 'str', '0.1'),
            ('properties', 'software.component_property', '0.N'),
            ('release_date', 'datetime', '0.1'),
            ('responsible_parties', 'shared.responsible_party', '0.N'),
            ('short_name', 'str', '1.1'),
            ('version', 'str', '0.1'),
        ],
        'doc_strings': {
            'code_access': 'Instructions on how to access the source code for this component.',
            'coupling_framework': 'The coupling framework that this entire component conforms to.',
            'description': 'A free-text description of the component.',
            'funding_sources': 'The entities that funded this software component.',
            'grid': 'A reference to the grid that is used by this component.',
            'is_embedded': 'An embedded component cannot exist on its own as an atomic piece of software; instead it is embedded within another (parent) component. When embedded equals "true", the SoftwareComponent has a corresponding piece of software (otherwise it is acting as a "virtual" component which may be inexorably nested within a piece of software along with several other virtual components).',
            'license': 'The license held by this piece of software.',
            'long_name': 'The name of the model (that is recognized externally).',
            'online_resource': 'Provides a URL location for this model.',
            'properties': 'The properties that this model simulates and/or couples.',
            'release_date': 'The date of publication of the software component code (as opposed to the date of publication of the metadata document, or the date of deployment of the model)',
            'short_name': 'The name of the model (that is used internally).',
            'version': 'A free text description of the component version #.'
        },
        'decodings' : [
            ('sub_components', 'child::cim:childComponent/cim:modelComponent', 'software.model_component'),
            ('sub_components', 'child::cim:childComponent/cim:processorComponent', 'software.processor_component'),
            ('citations', 'child::cim:citation'),
            ('description', 'child::cim:description'),
            ('language', 'child::cim:componentLanguage'),
            ('long_name', 'child::cim:longName'),
            ('properties', 'child::cim:componentProperties/cim:componentProperty'),
            ('properties', 'child::cim:scientificProperties/cim:componentProperty'),
            ('properties', 'child::cim:numericalProperties/cim:componentProperty'),
            ('release_date', 'child::cim:releaseDate'),
            ('responsible_parties', 'child::cim:responsibleParty'),
            ('short_name', 'child::cim:shortName'),
        ]
    }


def spatial_regridding():
    """Characteristics of the scheme used to interpolate a field from one grid (source grid) to another (target grid).  Documents should use either the spatialRegriddingStandardMethod _or_ the spatialRegriddingUserMethod, but not both.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('dimension', 'software.spatial_regridding_dimension_type', '0.1'),
            ('properties', 'software.spatial_regridding_property', '0.N'),
            ('standard_method', 'software.spatial_regridding_standard_method_type', '0.1'),
            ('user_method', 'software.spatial_regridding_user_method', '0.1'),
        ],
        'decodings' : [
            ('dimension', 'child::cim:spatialRegriddingDimension'),
            ('properties', 'child::cim:spatialRegriddingProperty'),
            ('standard_method', 'child::cim:spatialRegriddingStandardMethod'),
            ('user_method', 'child::cim:spatialRegriddingUserMethod'),
        ]
    }


def spatial_regridding_property():
    """Used for OASIS-specific regridding information (ie: masked, order, normalisation, etc.)

    """
    return {
        'type' : 'class',
        'base' : 'shared.property',
        'is_abstract' : False
    }


def spatial_regridding_user_method():
    """Characteristics of the scheme used to interpolate a field from one grid (source grid) to another (target grid).  Documents should use either the spatialRegriddingStandardMethod _or_ the spatialRegriddingUserMethod, but not both.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('file', 'data.data_object', '0.1'),
            ('name', 'str', '1.1')
        ],
        'decodings' : [
            ('file', 'child::cim:file/cim:dataObject'),
            ('file', 'child::cim:file/cim:reference', 'shared.doc_reference'),
            ('name', 'child::cim:name')
        ]
    }


def statistical_model_component():
    """Creates and returns instance of statistical_model_component class.

    """
    return {
        'type' : 'class',
        'base' : 'software.component',
        'is_abstract' : False,
        'is_document': True,
        'properties' : [
            ('type', 'software.statistical_model_component_type', '0.1'),
            ('types', 'software.statistical_model_component_type', '1.N'),
            ('timing', 'software.timing', '0.1'),
        ],
        'doc_strings': {
            'type': 'Describes the type of component. There can be multiple types.',
            'types': 'Describes the type of component. There can be multiple types.',
            'timing': 'Describes information about how this component simulates time.',
        },
        'decodings' : [
            ('meta', 'self::cim:statisticalModelComponent'),
            ('timing', 'child::cim:timing'),
            ('type', 'child::cim:type[1]/@value'),
            ('types', 'child::cim:type/@value'),
        ]
    }


def time_lag():
    """The coupling field used in the target at a given time corresponds to a field produced by the source at a previous time. This lag specifies the difference in time.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('value', 'int', '0.1'),
            ('units', 'software.timing_units', '0.1'),
        ],
        'decodings' : [
            ('value', 'child::cim:value'),
            ('units', '@units'),
        ]
    }


def timing():
    """Provides information about the rate of couplings and connections and/or the timing characteristics of individual components - for example, the start and stop times that the component was run for or the units of time that a component is able to model (in a single timestep).

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('end', 'datetime', '0.1'),
            ('is_variable_rate', 'bool', '0.1'),
            ('rate', 'int', '0.1'),
            ('start', 'datetime', '0.1'),
            ('units', 'software.timing_units', '0.1'),
        ],
        'doc_strings': {
            'is_variable_rate': 'Describes whether or not the model supports a variable timestep. If set to true, then rate should not be specified.'
        },
        'decodings' : [
            ('end', 'child::cim:end'),
            ('is_variable_rate', '@variableRate'),
            ('rate', 'child::cim:rate'),
            ('start', 'child::cim:start'),
            ('units', '@units'),
        ]
    }


def time_transformation():
    """The coupling field used in the target at a given time corresponds to a field produced by the source at a previous time. This lag specifies the difference in time.

    """
    return {
        'type' : 'class',
        'base' : None,
        'is_abstract' : False,
        'properties' : [
            ('description', 'str', '0.1'),
            ('mapping_type', 'software.time_mapping_type', '1.1'),
        ],
        'decodings' : [
            ('description', 'child::cim:description'),
            ('mapping_type', 'child::cim:mappingType/@value'),
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.codecs.json.__init__.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Python json codec.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc.cv.codecs.json.decoder import decode
from pyesdoc.cv.codecs.json.encoder import encode



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.model.term.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary term, e.g. IPSL.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyesdoc.cv
from pyesdoc.cv.model.entity import Entity



class Term(Entity):
    """A vocabulary term.

    """
    def __init__(self):
        """Instance constructor.

        """
        self.alternative_name = None    # primary synonym
        self.alternative_url = None     # an alternative URL
        self.associations = set()       # associated terms
        self.collection = None          # collection of which term is a member
        self.create_date = None         # date term was created
        self.data = None                # associated data
        self.description = None         # descriptive text
        self.io_path = None             # Path to I/O.
        self.idx = None                 # term integer based identifier
        self.io_path = None             # file system path
        self.label = None               # user interface label
        self.name = None                # canonical name
        self.parent = None              # parent term within collection
        self.status = None              # governance status
        self.synonyms = list()          # name synonyms
        self.uid = None                 # universally unique identifier
        self.url = None                 # a url for further info


    def __repr__(self):
        """Instance representation.

        """
        return u"{} -> [{}]".format(self.namespace, self.status)


    def __contains__(self, key):
        """Instance membership predicate.

        """
        key = unicode(key).strip().lower()

        return key in self.all_names


    @property
    def authority(self):
        """Gets associated governing authority.

        """
        return self.scope.authority


    @property
    def scope(self):
        """Gets associated scope.

        """
        return self.collection.scope


    @property
    def namespace(self):
        """Gets namespace.

        """
        return ":".join([
            self.collection.namespace,
            self.name
            ])


    @property
    def full_idx(self):
        """Gets full computed idx.

        """
        return u"{}.{}".format(
            self.collection.full_idx,
            self.idx
            )


    @property
    def all_names(self):
        """Returns all term names.

        """
        result = [self.name, self.alternative_name] + self.synonyms
        result = [t for t in result if t is not None and len(t) > 0]

        return set(sorted(result))


    @property
    def depth(self):
        """Gets hierarchical depth.

        """
        return len(self.ancestors)


    @property
    def ancestors(self):
        """Gets ancestral hierarchy.

        """
        result = []
        ancestor = self.parent
        while ancestor:
            result.append(ancestor)
            ancestor = ancestor.parent
        result.reverse()

        return result


    @property
    def sort_key(self):
        """Gets the term's sort key

        """
        return u"{}::{}".format(self.namespace, self.name).lower()


    def add_synonym(self, new_synonym):
        """Adds a new synonym to the term's synnym set.

        """
        # Validate new synonym.
        _validate_term_synonym(new_synonym)

        # Format.
        new_synonym = unicode(new_synonym).strip()

        # Escape if already aliased.
        for synonym in self.synonyms:
            if synonym == new_synonym:
                return

        # Update synonym set.
        self.synonyms = sorted(self.synonyms + [new_synonym])

        # Save term.
        # self.partition.save(self)


    @property
    def is_cached(self):
        """Returns flag indicating whether the term is cached or not.

        """
        return self.partition.is_cached(self)


    @property
    def is_written(self):
        """Returns flag indicating whether the term is written to file system or not.

        """
        return self.partition.is_written(self)


    @property
    def partition(self):
        """Returns associated partition.

        """
        return pyesdoc.cv.get_partition(self.domain)


    def accept(self):
        """Marks as accepted.

        """
        self.partition.accept(self)


    def reject(self):
        """Marks as rejected.

        """
        self.partition.reject(self)


    def save(self):
        """Saves term to persistant state stores.

        """
        self.partition.save(self)


    def deprecate(self):
        """Marks as deprecated.

        """
        self.partition.deprecate(self)


    def destroy(self):
        """Destroys term from all persistant state stores.

        """
        self.partition.destroy(self)


    def associate(self, term):
        """Appends an associated term to managed collection.

        :param pyesdoc.cv.Term term: Associated term to be added.

        """
        self.associations.add(term)



# -*- coding: utf-8 -*-

"""
.. module:: dict.decoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Decodes a term from a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import arrow
import uuid

from pyesdoc.cv.model import Term
from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Scope
from pyesdoc.cv.model import Collection



def decode(obj):
    """Decodes a term from a dictionary.

    :param dict obj: Dictionary to be decoded.

    :returns: Decoded term.
    :rtype: pyesdoc.cv.Term

    """
    try:
        _DECODERS[obj['_type']]
    except KeyError:
        raise TypeError("Decoding type unsupported: {}".format(obj['_type']))

    return _DECODERS[obj['_type']](obj)


def _decode_authority(obj):
    """Decodes a termset from a dictionary.

    """
    instance = Authority()
    instance.description = obj['description']
    instance.label = obj['label']
    instance.name = obj['name']
    instance.scopes = [_decode_scope(i) if isinstance(i, dict) else i
                       for i in obj['scopes']]
    instance.url = obj['url']

    # Wire hierarchy.
    for scope in instance:
        if isinstance(scope, Scope):
            scope.authority = instance

    return instance


def _decode_scope(obj):
    """Decodes a termset from a dictionary.

    """
    instance = Scope()
    instance.collections = [_decode_collection(i) if isinstance(i, dict) else i
                            for i in obj['collections']]
    instance.description = obj['description']
    instance.idx = obj['idx']
    instance.label = obj['label']
    instance.name = obj['name']
    instance.uid = uuid.UUID(unicode(obj['uid']))
    instance.url = obj['url']

    # Wire hierarchy.
    for collection in instance:
        if isinstance(collection, Collection):
            collection.scope = instance

    return instance


def _decode_collection(obj):
    """Decodes a termset from a dictionary.

    """
    instance = Collection()
    instance.create_date = arrow.get(obj['create_date']).datetime
    instance.description = obj['description']
    instance.idx = obj['idx']
    instance.label = obj['label']
    instance.name = obj['name']
    instance.terms = [_decode_term(i) if isinstance(i, dict) else i
                      for i in obj['terms']]
    instance.uid = uuid.UUID(unicode(obj['uid']))
    instance.url = obj['url']

    # Wire hierarchy.
    for term in instance:
        if isinstance(term, Term):
            term.collection = instance

    return instance


def _decode_term(obj):
    """Decodes a term from a dictionary.

    """
    instance = Term()
    instance.alternative_name = obj['alternative_name']
    instance.alternative_url = obj['alternative_url']
    instance.create_date = arrow.get(obj['create_date']).datetime
    instance.data = obj['data']
    instance.description = obj['description']
    instance.idx = obj['idx']
    instance.label = obj['label']
    instance.name = obj['name']
    instance.status = obj['status']
    instance.synonyms = obj['synonyms']
    instance.uid = uuid.UUID(unicode(obj['uid']))
    instance.url = obj['url']
    if instance.parent:
        instance.parent = uuid.UUID(unicode(obj['parent']))

    return instance


# Map of supported type keys to decoding functions.
_DECODERS = {
    Authority.__module__: _decode_authority,
    Collection.__module__: _decode_collection,
    Scope.__module__: _decode_scope,
    Term.__module__: _decode_term
}



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc._validation.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Validates a document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import re

from pyesdoc.ontologies.graph_parser import GraphParser
from pyesdoc import ontologies
from pyesdoc import constants



# Validation error text.
_ERR_CONSTANT = "constant value has been overwritten"
_ERR_LIST_IS_EMPTY = "is an empty list"
_ERR_LIST_CONTAINS_INVALID_ITEM_TYPES = "list contains item(s) of invalid type"
_ERR_LIST_TYPE = "is of invalid type (actual = {0}, expected=list)"
_ERR_HIDDEN = "hidden property has been assigned"
_ERR_ITEM_IS_NULL = "is null"
_ERR_ITEM_TYPE = "is of invalid type (actual = {0}, expected={1})"
_ERR_ITEM_IS_EMPTY_TEXT = "is zero length"
_ERR_ITEM_IS_LIST = "is a list, expected a complex or ximple type"
_ERR_REGEX = "regular expression validation failed"

# Type of document references.
_DOCUMENT_REFERENCE_TYPE = "DocReference"

# Set of text types.
_TEXT_TYPES = (str, unicode)


def _validate_cardinality(value, expected_cardinality):
    """Cardinality constraint validator.

    """
    if expected_cardinality == constants.CARDINALITY_TYPE_0_0:
        if value not in (None, []):
            return _ERR_HIDDEN

    elif expected_cardinality == constants.CARDINALITY_TYPE_0_1:
        if isinstance(value, list):
            return _ERR_ITEM_IS_LIST

    elif expected_cardinality == constants.CARDINALITY_TYPE_1_1:
        if isinstance(value, list):
            return _ERR_ITEM_IS_LIST
        if value is None:
            return _ERR_ITEM_IS_NULL

    elif expected_cardinality == constants.CARDINALITY_TYPE_1_N:
        if not isinstance(value, list):
            return _ERR_LIST_TYPE.format(type(value))
        if len(value) == 0:
            return _ERR_LIST_IS_EMPTY

    elif expected_cardinality == constants.CARDINALITY_TYPE_0_N:
        if not isinstance(value, list):
            return _ERR_LIST_TYPE.format(type(value))


def _validate_constant(value, expected_constant):
    """Constant constraint validator.

    """
    if not expected_constant:
        return

    elif value != expected_constant:
        return _ERR_CONSTANT


def _validate_regex(value, regex):
    """Regular expression constraint validator.

    """
    if not regex:
        return

    elif not re.match(regex, str(value)):
        return _ERR_REGEX


def _validate_type(value, expected_type):
    """Type constraint validator.

    """
    def _is_type_mismatch(instance):
        """Validates that a complex type matches.

        """
        if isinstance(instance, expected_type):
            return False

        if expected_type in ontologies.type_info.TYPES and \
           type(instance).__name__.split(".")[-1] == _DOCUMENT_REFERENCE_TYPE:
            return False

        return True


    # Escape if unncessary.
    if value in (None, []):
        return

    # Validate collections.
    if isinstance(value, list):
        if [i for i in value if _is_type_mismatch(i)]:
            return _ERR_LIST_CONTAINS_INVALID_ITEM_TYPES

    # Assert non text.
    elif expected_type not in _TEXT_TYPES:
        if _is_type_mismatch(value):
            return _ERR_ITEM_TYPE.format(type(value), expected_type)

    # Assert text.
    elif not isinstance(value, _TEXT_TYPES):
        return _ERR_ITEM_TYPE.format(type(value), expected_type)

    # Assert text length.
    elif len(value.strip()) == 0:
        return _ERR_ITEM_IS_EMPTY_TEXT


# Set of property validators to be executed.
_VALIDATORS = (
    (constants.CONSTRAINT_TYPE_CARDINALITY, _validate_cardinality),
    (constants.CONSTRAINT_TYPE_TYPEOF, _validate_type),
    (constants.CONSTRAINT_TYPE_CONSTANT, _validate_constant),
    (constants.CONSTRAINT_TYPE_REGEX, _validate_regex),
    )


class _GraphParserValidator(GraphParser):
    """A graph parser that performs validation.

    """
    def __init__(self):
        """Instance initializer.

        """
        self.errors = []


    def on_property_parse(self, container, path, name, value):
        """Property parsed event handler.

        """
        for constraint_type, validator in _VALIDATORS:
            error = validator(value, ontologies.get_property_constraint(container, name, constraint_type))
            if error:
                self.errors.append("doc{} --> {}".format(path, error))
                return


def validate(doc):
    """Validates a document.

    :param object doc: A pyesdoc document instance.

    :returns: A list of validation errors.
    :rtype: list

    """
    parser = _GraphParserValidator()
    parser.parse(doc)

    return parser.errors


def is_valid(doc):
    """Returns validation status of a document.

    :param object doc: A pyesdoc document instance.

    :returns: The document's validation state.
    :rtype: bool

    """
    return len(validate(doc)) == 0




# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.utils.runtime.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Runtime utility functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Set of logging levels.
LOG_LEVEL_DEBUG = 'DUBUG'
LOG_LEVEL_INFO = 'INFO'
LOG_LEVEL_WARNING = 'WARNING'
LOG_LEVEL_ERROR = 'ERROR'


def assert_doc(doc):
    """Asserts thay passed variable is a pyesdoc object instance.

    :param object doc: A pyesdoc object instance.

    """
    if doc is None or not isinstance(doc, object):
        raise TypeError("Document type is unrecognized")
    if not hasattr(doc, 'meta'):
        raise ValueError('Document has no meta attribute')


def log(msg=None, level=LOG_LEVEL_INFO):
    """Outputs a message to log.

    :param str msg: Message for writing to log.
    :param str level: Logging level.

    """
    # Format.
    if msg:
        msg = "ES-DOC :: pyesdoc :: {0} > {1}".format(level, str(msg).strip())
    else:
        msg = '-----------------------------------------'
        msg += msg

    # TODO output to logs.
    print msg


def log_error(err):
    """Logs a runtime error.

    :param Exception err: Exception to be logged.

    """
    if isinstance(err, basestring):
        msg = "!!! RUNTIME ERROR !!! :: {}.".format(err)
    else:
        msg = "!!! RUNTIME ERROR !!! :: {} :: {}.".format(err.__class__, err)
    log(msg, level=LOG_LEVEL_ERROR)


def log_warning(msg):
    """Logs a runtime warning.

    :param str msg: Message for writing to log.

    """
    log(msg, level=LOG_LEVEL_WARNING)


def invoke(ctx, actions, error_actions):
    """Invokes a set of actions.

    :param object ctx: Processing context information.
    :param iterable actions: Set of actions to perform.
    :param iterable error_actions: Set of error actions to perform.

    """
    try:
        for action in actions:
            action(ctx)
    except Exception as exc:
        ctx.error = exc
        for action in error_actions:
            action(ctx)


def first(collection, name, value, value_formatter=None):
    """Returns first item in collection filtered by attribute name/value.

    :param iterable collection: A collection of items for filtering.
    :param str name: Name of an attribute acting as filter target.
    :param object value: Value to filter by.
    :param function formatter: Value formatting function.

    :returns: First item in collection filtered by attribute name/value.
    :rtype: object | None

    """
    filtered = all(collection, name, value, value_formatter)

    return filtered[0] if len(filtered) else None


def all(collection, name, value, formatter=None):
    """Returns collection filtered by attribute name/value.

    :param iterable collection: A collection of items for filtering.
    :param str name: Name of an attribute acting as filter target.
    :param object value: Value to filter by.
    :param function formatter: Value formatting function.

    :returns: Collection filtered by attribute name/value.
    :rtype: iterable

    """
    if formatter:
        value = formatter(value)

    def predicate(i):
        if hasattr(i, name):
            v = getattr(i, name)
            if formatter:
                v = formatter(v)

            return v == value
        else:
            return False

    return filter(predicate, collection)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.cv.validation.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates domain model class instance validation.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.cv.constants import REGEX_CANONICAL_NAME
from pyesdoc.cv.model import TYPES
from pyesdoc.cv.model import TYPE_INFO



def validate(instance):
    """Validates an instance of a domain model class.

    :param instance: Sub-class of pyesdoc.cv.Entity

    :returns: Set of instance validation errrors.
    :rtype: set

    """
    if not isinstance(instance, TYPES):
        raise NotImplementedError("Invalid instance: unknown type")

    return _validate_instance(instance)


def is_valid(instance):
    """Gets flag indicating validity status of a domain model class.

    :returns: Validity status of a domain model class.
    :rtype: bool

    """
    return len(validate(instance)) == 0


def get_errors(instance):
    """Returns sorted list of domain model class instance validation errors.

    :returns: Sorted list of domain model class instance validation errors.
    :rtype: list

    """
    return sorted(list(validate(instance)))


def _validate_instance(instance):
    """Validates an instance.

    """
    errs = set()

    for field_info in TYPE_INFO[type(instance)]:
        try:
            _validate_field(instance, field_info)
        except ValueError as err:
            errs.add("{}.{}".format(instance.__class__.__name__, err.message))

    return errs


def _validate_field(instance, type_info):
    """Validates a field.

    """
    # Unpack type info.
    try:
        field, typeof, cardinality, misc = type_info
    except ValueError:
        field, typeof, cardinality = type_info
        misc = None
    if isinstance(misc, unicode):
        pass
        # print "TODO: validate {}.{} by regex".format(instance.__class__.__name__, field)

    # Error: unknown field.
    if not hasattr(instance, field):
        raise ValueError("{}: unknown".format(field))

    # Set validation function.
    if cardinality in {"0.N", "1.N"}:
        func = _validate_iterable
    else:
        func = _validate_value

    # Execute validation function.
    val = getattr(instance, field)
    is_mandatory = (cardinality in {"1.1", "1.N"})
    try:
        func(val, is_mandatory, typeof, misc)
    except ValueError as err:
        raise ValueError("{}: {}".format(field, err))


def _validate_value(val, is_mandatory, typeof, misc):
    """Validates an iterable field value.

    """
    # Error: mandatory.
    if is_mandatory and val is None:
        raise ValueError("undefined")

    # Error: type mismatch.
    if val is not None and not isinstance(val, typeof):
        raise ValueError("type mismatch")

    # Error: unicode length.
    # TODO

    # Error: enum.
    if isinstance(misc, tuple) and val not in misc:
        raise ValueError("not in enum")

    # Error: regex.
    if isinstance(misc, unicode):
        # TODO: validate regex
        pass


def _validate_iterable(val, is_mandatory, typeof, misc):
    """Validates an iterable field value.

    """
    # Error: type mismatch.
    if not isinstance(val, list):
        raise ValueError("type")

    # Error: length.
    if is_mandatory and len(val) == 0:
        raise ValueError("undefined")

    # Error: item type mismatch.
    if [i for i in val if not isinstance(i, typeof)]:
        raise ValueError("item type mismatch")

    # Error: enum.
    if isinstance(misc, tuple) and [i for i in val if i not in misc]:
        raise ValueError("item not in enum")

    # Error: regex.
    if isinstance(misc, unicode):
        # TODO: validate regex
        pass


def validate_authority_description(description):
    """Validates a termset's description.

    """
    validate_unicode(description, "description")


def validate_authority_name(name):
    """Validates a term's name.

    """
    validate_unicode(name, "name")
    for char in REGEX_CANONICAL_NAME:
        if name.find(char) > -1:
            raise ValueError('invalid name')


def validate_authority_url(url):
    """Validates a term's url.

    """
    validate_url(url, "url")


def validate_collection_description(description):
    """Validates a termset's description.

    """
    validate_unicode(description, "description")


def validate_collection_name(name):
    """Validates a term's name.

    """
    validate_unicode(name, "name")
    for char in REGEX_CANONICAL_NAME:
        if name.find(char) > -1:
            raise ValueError('invalid name')


def validate_scope_description(description):
    """Validates a termset's description.

    """
    validate_unicode(description, "description")


def validate_scope_name(name):
    """Validates a term's name.

    """
    validate_unicode(name, "name")
    for char in REGEX_CANONICAL_NAME:
        if name.find(char) > -1:
            raise ValueError('invalid name')


def validate_scope_url(url):
    """Validates a term's url.

    """
    validate_url(url, "url")


def validate_term_data(data):
    """Validates a term's data.

    """
    if not isinstance(data, dict):
        raise ValueError("invalid term data")


def validate_term_name(name):
    """Validates a term's name.

    """
    validate_unicode(name, "name")
    for char in REGEX_CANONICAL_NAME:
        if name.find(char) > -1:
            raise ValueError('invalid name')


def validate_unicode(val, field):
    """Validates a unicode field value.

    """
    if val is None:
        raise ValueError("undefined {}".format(field))
    if not isinstance(val, unicode):
        raise ValueError("invalid {} (unicode test failed)".format(field))
    if not len(val.strip()):
        raise ValueError("invalid {} (>0 length test failed)".format(field))


def validate_url(val, field):
    """Validates a url field value.

    """
    validate_unicode(val, field)
    # TODO: apply regex



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.definition.initializer.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates definition package initialization.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyesdoc
from pyesdoc.drq.definition.wrapper import Wrapper
from pyesdoc.drq.definition.table import Table
from pyesdoc.drq.definition.table_attribute import TableAttribute
from pyesdoc.drq.definition.xml_parser import XMLParser



class _Parser(XMLParser):
    """Initialises library with parsed definition.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(_Parser, self).__init__()
        self.tables = {}


    def on_parse_table_element(self, elem):
        """On table element parse event handler.

        """
        self.tables[elem] = Table(elem)


    def on_parse_table_attribute_element(self, table_elem, elem):
        """On table row attribute element parse event handler.

        """
        table = self.tables[table_elem]
        attribute = TableAttribute(table, elem)
        duplicates = [i for i in table if i.label == attribute.label]
        if not duplicates:
            table.attributes.append(attribute)


    def on_parse_complete(self):
        """On parse complete event handler.

        """
        # Instantiate wrapper.
        wrapper = Wrapper(self.tables.values())

        # Mutate drq.definition from package to wrapper.
        setattr(pyesdoc.drq, "definition", wrapper)


def execute():
    """Execute initialization process.

    """
    parser = _Parser()
    parser.execute()



from pyesdoc.cv.model import Authority
from pyesdoc.cv.model import Term
from pyesdoc.cv.io_mgr.delete import delete_authority
from pyesdoc.cv.io_mgr.delete import delete_term
from pyesdoc.cv.io_mgr.read import read_authority
from pyesdoc.cv.io_mgr.write import write_authority




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.definition.wrapper.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps dreq2Defn.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class Wrapper(object):
    """Wraps dreq2Defn.xml.

    """
    def __init__(self, tables):
        """Instance constructor.

        :param list tables: Collection of definition tables.

        """
        self._tables = sorted(tables, key=lambda i: i.label.lower())
        for table in tables:
            setattr(self, str(table), table)


    def __repr__(self):
        """Instance representation.

        """
        return unicode(self._tables)


    def __len__(self):
        """Returns number of attributes in managed collection.

        """
        return len(self._tables)


    def __contains__(self, key):
        """Returns flag indicating whether a certain table exists or not.

        """
        if key in self._tables:
            return True
        return self[key] is not None


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(self._tables)


    def __getitem__(self, key):
        """Returns a child table attribute.

        """
        try:
            int(key)
        except ValueError:
            key = str(key).strip().lower()
            for table in self._tables:
                if table.label.lower() == key or \
                   table.label_drq.lower() == key:
                    return table
        else:
            return self._tables[key]




# -*- coding: utf-8 -*-

"""
.. module:: test_utils_assert.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes test assertion utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import datetime
import inspect
import os



# Integer assertion constants.
COMPARE_EXACT = "EXACT"
COMPARE_GT = "GT"
COMPARE_GTE = "GTE"
COMPARE_LTE = "LTE"
COMPARE_LT = "LT"
COMPARE_TYPES = (
    COMPARE_EXACT,
    COMPARE_GT,
    COMPARE_GTE,
    COMPARE_LT,
    COMPARE_LTE
)


def assert_iter(collection,
                length=-1,
                item_type=None,
                length_compare=COMPARE_EXACT):
    """Asserts an object collection.

    :param list collection: An object collection.
    :param int length: Collection size.
    :param str length: Collection size comparason operator.
    :param class item_type: Type that each collection item should sub-class.

    """
    assert_object(collection)
    assert iter(collection) is not None
    if length != -1:
        assert_int(len(collection), length, length_compare)
    if item_type is not None:
        if isinstance(collection, dict):
            collection = collection.values()
        for instance in collection:
            assert_object(instance, item_type)


def assert_in_collection(collection, item_attr, items):
    """Asserts an item is within a collection.

    :param sequence collection: A collection.
    :param str item_attr: Attribute name upon from which to derive assertion collection.
    :param object items: Set of items to assert whether found in collection.

    """
    try:
        iter(items)
    except TypeError:
        items = {items}
    targets = None
    if item_attr is not None:
        targets = [getattr(i, item_attr) for i in collection]
    else:
        targets = collection
    for item in items:
        assert item in targets, item


def assert_has_member(mod, member):
    """Asserts that a module exposes a member.

    :param module mod: Module that should expose the passed member.
    :param str member: Name of a function, exception, class ..etc.

    """
    assert inspect.ismodule(mod)
    assert hasattr(mod, member), "Missing member: {}".format(member)


def assert_has_class(mod, cls):
    """Asserts that a container exposes a class.

    :param module mod: Container module that should expose the passed class.
    :param str cls: Name of a class.

    """
    assert_has_member(mod, cls)
    assert inspect.isclass(getattr(mod, cls)), \
           "{} is not a class".format(cls)


def assert_has_constant(mod, constant):
    """Asserts that a container exposes a constant.

    :param module mod: Container module that should expose the passed constant.
    :param str constant: Name of a constant.

    """
    assert_has_member(mod, constant)


def assert_has_function(mod, func):
    """Asserts that a container exposes a function.

    :param module mod: Container module that should expose the passed function.
    :param str func: Name of a function.

    """
    assert_has_member(mod, func)
    assert inspect.isfunction(getattr(mod, func)), \
           "{} is not a function".format(func)


def assert_has_exception(mod, err):
    """Asserts that a container exposes an exception.

    :param module mod: Container module that should expose an exception class.
    :param str err: Name of an exception class.

    """
    assert_has_class(mod, err)
    assert issubclass(getattr(mod, err), Exception), \
           "Exception type does not inherit from builtin Exception class."


def assert_none(instance):
    """Asserts an instance is none.

    :param object instance: An object for testing.

    """
    assert instance is None, \
           "Instance null mismatch : actual = {0} - {1} :: expected = None" \
           .format(type(instance), instance)


def assert_object(instance, instance_type=None):
    """Asserts an object instance.

    :param object instance: An object for testing.
    :param class instance_type: Type that object must either be or sub-class from.

    """
    assert instance is not None, "Instance is none"
    if instance_type is not None:
        assert isinstance(instance, instance_type), \
               "Instance type mismatch : actual = {0} :: expected = {1}" \
               .format(type(instance), instance_type)


def assert_objects(instance1, instance2, instance_type=None):
    """Asserts that 2 object instances are equal.

    :param object instance1: An object for testing.
    :param object instance2: An object for testing.

    """
    assert_object(instance1, instance_type)
    assert_object(instance2, instance_type)
    assert instance1 == instance2, "Instances are not equal"


def assert_bool(actual, expected):
    """Asserts a boolean.

    :param str actual: Actual boolean value.
    :param str expected: Expected boolean value.

    """
    assert bool(actual) == bool(expected)


def assert_str(actual, expected, startswith=False):
    """Asserts a string.

    :param str actual: Actual string value.
    :param str expected: Expected string value.

    :param startswith: Flag indicating whether to perform startswith test.
    :type startswith: bool

    """
    # Format.
    actual = str(actual).strip()
    expected = str(expected).strip()

    # Assert.
    if startswith == False:
        assert actual == expected, \
               "String mismatch : actual = {0} :: expected = {1}" \
               .format(actual, expected)
    else:
        assert actual.startswith(expected) == True, \
               "String startswith mismatch : actual = {0} :: expected = {1}" \
               .format(actual, expected)


def assert_unicode(actual, expected):
    """Asserts a unicode.

    :param str actual: Actual unicode value.
    :param str expected: Expected unicode value.

    """
    assert_object(actual, unicode)
    assert_object(expected, unicode)
    assert actual == expected, \
           "Unicode mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_date(actual, expected):
    """Asserts a datetime.

    :param str actual: Actual date value.
    :param str expected: Expected date value.

    """
    if not isinstance(actual, datetime.datetime):
        actual = get_date(actual)
    if not isinstance(expected, datetime.datetime):
        expected = get_date(expected)

    assert actual == expected, \
           "Date mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_float(actual, expected):
    """Asserts a float.

    :param float actual: Actual float value.
    :param float expected: Expected float value.

    """
    assert_object(actual, float)
    assert_object(expected, float)
    assert actual == expected, \
           "Float mismatch : actual = {0} :: expected = {1}" \
           .format(actual, expected)


def assert_path(actual):
    """Asserts a filepath.

    :param str actual: Actual file path.

    """
    assert_bool(os.path.exists(actual), True)


def assert_int(actual, expected, assert_type=COMPARE_EXACT, msg=None):
    """Asserts an integer.

    :param str actual: Actual integer value.
    :param str expected: Expected integer value.

    """
    # Parse actual value.
    # ... convert string
    if type(actual) == str:
        actual = int(actual)
    # ... collection length checks
    else:
        try:
            iter(actual)
        except TypeError:
            pass
        else:
            actual = len(actual)

    if assert_type == COMPARE_EXACT:
        assert expected == actual, \
               "{0} != {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_GT:
        assert expected > actual, \
               "{0} !> {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_GTE:
        assert expected >= actual, \
               "{0} !>= {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_LT:
        assert expected < actual, \
               "{0} !< {1} {2}".format(actual, expected, msg)
    elif assert_type == COMPARE_LTE:
        assert expected <= actual, \
               "{0} !<= {1} {2}".format(actual, expected, msg)
    else:
        assert expected == actual, \
               "{0} != {1} {2}".format(actual, expected, msg)


def assert_int_negative(actual, expected):
    """Negatively asserts an integer.

    :param str actual: Actual integer value.
    :param str expected: Expected integer value.

    """
    assert actual != expected


def assert_uuid(actual, expected):
    """Asserts a uuid.

    :param str actual: Actual uuid value.
    :param str expected: Expected uuid value.

    """
    if isinstance(actual, uuid.UUID) == False:
        actual = uuid.UUID(actual)
    if isinstance(expected, uuid.UUID) == False:
        expected = uuid.UUID(expected)

    assert actual == expected, "{0} != {1}".format(actual, expected)



# -*- coding: utf-8 -*-

"""Encapsualtes a set of qxml specific name conversion operations.

"""
import os

from pyesdoc.mp.ontologies.generators.generator_utils import *



# Language name constant.
LANGUAGE = 'qxml'

# Language file extension constant.
FILE_EXTENSION = '.xml'

# Set of simple type mappings.
_SIMPLE_TYPE_MAPPINGS = {
    # TODO: THE Q ALSO INCLUDES 'EMAIL' & 'TIME'
    # TODO: CIM2 IS SUPPOSED TO INCLUDE 'email'
    # TODO: CHECK THE STATUS OF THIS
    'bool' : 'BOOLEAN',
    'date' : 'DATE',
    'datetime' : 'DATETIME',
    'float' : 'DECIMAL',
    'int' : 'INTEGER',
    'str' : 'STRING',
    'unicode' : 'STRING',
    'uri' : 'URL',
    'uuid' : 'STRING',
}

QXML_ATOMIC_TYPE = "ATOMIC"
QXML_ENUMERATION_TYPE = "ENUMERATION"
QXML_RELATIONSHIP_TYPE = "RELATIONSHIP"


def _strip(name):
    """Returns stripped name.

    Keyword Arguments:
    name - name being converted.

    """
    if isinstance(name, str) == False:
        name = name.name
    return name


def _strip_class_name(name):
    """Returns stripped class name.

    Keyword Arguments:
    name - name being converted.

    """
    name = _strip(name)
    if name.find('.') != -1:
        name = name.split('.')[len(name.split('.')) - 1]
    return name


def get_ontology_name(name):
    """Converts name to an ontology name suitable for QXML.

    Keyword Arguments:
    name - name being converted.

    """
    if isinstance(name, str) == False:
        name = name.name

    return name.lower()


def get_ontology_version(name):
    """Converts version identifier to a version suitable for QXML.

    Keyword Arguments:
    name - name of version identifier being converted.

    """
    if isinstance(name, str) == False:
        name = name.version

    return name


def _get_ontology_directory(ctx, include_version=True):
    """
    Returns ontology directory into which code is generated.

    :param GeneratorContext ctx: Generation context information.
    :param bool include_version: should version info be included in the directory

    """
    _dir = ''
    if ctx.output_dir is not None:
        _dir += ctx.output_dir + '/'
    _dir += get_ontology_name(ctx.ontology)
    if include_version:
        _dir += '/'
        _dir += 'v' + get_ontology_version(ctx.ontology)
    return _dir


def _get_ontology_filename(ctx, include_version=True):
    """
    Returns ontology filename for generated code file

    :param GeneratorContext ctx: Generation context information.
    :param bool include_version: should version info be included in the filename
    :return:
    """
    ontology = ctx.ontology
    ontology_filename = get_ontology_name(ontology)
    if include_version:
        ontology_filename += "_{0}".format(get_ontology_version(ontology).replace('.', '_'))
    ontology_filename += ".xml"
    return ontology_filename


def get_ontology_path(ctx):
    """
    Returns full ontology path for generated QXML file

    :param GeneratorContext ctx: Generation context information.
    :return:
    """
    ontology_path = os.path.join(
        _get_ontology_directory(ctx, include_version=True),
        _get_ontology_filename(ctx, include_version=True)
    )
    return ontology_path


def get_package_name(name):
    """Converts name to a package name suitable for QXML.

    :param str name: Package name being converted.

    """
    return _strip_package_name(name)


def get_full_class_name(c):
    """Converts name to a qualified class name.
    (This is only used for printing info via "format" below)

    :param str name: Class name being converted.

    """
    return get_package_name(c.package) + "." + get_class_name(c)


def get_class_name(name):
    """Converts name to a python class name.

    :param str name: Class name being converted.

    """
    return convert_to_camel_case(_strip_class_name(name))


def get_class_doc_string_name(name):
    """Converts name to one suitable for use in QXML documentation.

    :param str name: Class name being converted.

    """
    name = _strip_class_name(name)
    return name.replace('_', ' ')


def _get_class_base_name(c):
    """Converts name to a python base class name.

    """
    if c.base is None:
        return 'object'
    elif c.base.package == c.package:
        return get_class_name(c.base)
    else:
        return get_full_class_name(c.base)


def _strip_package_name(name):
    """Returns stripped package name.

    """
    name = _strip(name)
    if name.find('.') != -1:
        name = name.split('.')[0]
    return name


def get_qualified_enum_name(enum):
    """
    Returns qualified (ie: package + enum name) enumeration name.

    """
    if enum.package:
        return "{0}.{1}".format(enum.package, enum.name)
    else:
        return enum.name

def get_atomic_property_type(property):
    """
    returns the corresponding QXML atomic property type
    :param property: property
    :return: string
    """
    property_type = property.type
    property_type_name = property_type.name_of_type
    if property_type_name not in _SIMPLE_TYPE_MAPPINGS:
        # TODO: REMOVE THIS ONCE I'M SURE EVERYTHING WORKS
        raise ValueError("Unable to locate '{0}' in ATOMIC TYPE MAP for {1}".format(property_type, property))
    return _SIMPLE_TYPE_MAPPINGS[property_type_name]


def get_relationship_property_target(property):
    """
    Returns a QXML compatible name of the class that this property points to
    :param property:
    :return:
    """
    property_type = property.type
    property_type_name = property_type.name_of_type
    return property_type_name


def get_relationship_property_target_classes(property):
    """
    Returns a list of all classes this property can point to
    (It is a list instead of a single instance b/c of the possibility of pointing to classes which have children)
    :param property:
    :return:
    """
    property_type = property.type
    property_target_class = property_type.cls
    assert property_target_class is not None

    property_target_classes = []
    if not property_target_class.is_abstract:
        property_target_classes.append(property_target_class)

    concrete_classes = [cls for cls in property_type.ontology.classes if not cls.is_abstract]
    property_target_classes += [
        cls for cls in concrete_classes
        if property_target_class in cls.bases
    ]

    return property_target_classes


def get_property_type(property):
    """
    Returns the corresponding QXML property type
    :param property:
    :return:
    """
    property_type = property.type
    if property_type.is_class:
        return QXML_RELATIONSHIP_TYPE
    elif property_type.is_enum:
        return QXML_ENUMERATION_TYPE
    else:
        return QXML_ATOMIC_TYPE


def clean_xml_text(text):
    """
    escapes any non-xml-compatible text prior to adding it to an lxml element node
    (as per: http://stackoverflow.com/a/4181222)
    :param text: input text to clean
    :return: cleaned text to return
    """
    return text.decode('utf-8')


def format(o):
    """
    Configure ontology names.

    """
    o.op_name = get_ontology_name(o)
    o.op_version = get_ontology_version(o)

    for p in o.packages:
        p.op_name = get_package_name(p)

    for c in o.classes:
        c.op_base_name = _get_class_base_name(c)
        c.op_doc_string_name = get_class_doc_string_name(c)
        c.op_name = get_class_name(c)
        c.op_full_name = get_full_class_name(c)
